import re
from typing import Dict, Optional, Type, Union

from . import xdr as stellar_xdr
from .exceptions import AssetCodeInvalidError, AssetIssuerInvalidError
from .keypair import Keypair
from .strkey import StrKey

__all__ = ["Asset"]


class Asset:
    """The :class:`Asset` object, which represents an asset and its
    corresponding issuer on the Stellar network.

    The following example shows how to create an Asset object::

        from stellar_sdk import Asset

        native_asset = Asset.native()  # You can also create a native asset through Asset("XLM").
        credit_alphanum4_asset = Asset("USD", "GBSKJPM2FM6O2C6GVZNAUAMGXZ6I4QIUPMNWVDN2NZULPWWTV3GI2SOX")
        credit_alphanum12_asset = Asset("BANANA", "GA6VT2PDD73TNNRYLPJPJYAAI7EGKBATZ7V562S7XY7TJD4GNOXRG6OS")
        print(f"Asset type: {credit_alphanum4_asset.type}\\n"
              f"Asset code: {credit_alphanum4_asset.code}\\n"
              f"Asset issuer: {credit_alphanum4_asset.issuer}\\n"
              f"Is native asset: {credit_alphanum4_asset.is_native()}")

    For more information about the formats used for asset codes and how issuers
    work on Stellar's network, see `Stellar's guide on assets`_.

    :param code: The asset code, in the formats specified in `Stellar's
        guide on assets`_.
    :param issuer: The account ID of the issuer. Note if the
        currency is the native currency (XLM (Lumens)), no issuer is necessary.
    :raises:
        | :exc:`AssetCodeInvalidError <stellar_sdk.exceptions.AssetCodeInvalidError>`: if ``code`` is invalid.
        | :exc:`AssetIssuerInvalidError <stellar_sdk.exceptions.AssetIssuerInvalidError>`: if ``issuer`` is not a valid ed25519 public key.

    .. _Stellar's guide on assets:
        https://developers.stellar.org/docs/glossary/assets/
    """

    def __init__(self, code: str, issuer: Optional[str] = None) -> None:
        Asset.check_if_asset_code_is_valid(code)

        if code != "XLM" and issuer is None:
            raise AssetIssuerInvalidError(
                "The issuer cannot be `None` except for the native asset."
            )

        if issuer is not None and not StrKey.is_valid_ed25519_public_key(issuer):
            raise AssetIssuerInvalidError("The issuer should be a correct public key.")

        self.code: str = code
        self.issuer: Optional[str] = issuer
        self._type: str = self.guess_asset_type()

    @staticmethod
    def check_if_asset_code_is_valid(code: str) -> None:
        """Check whether the `code` passed in by the user is a valid asset code,
        if not, an exception will be thrown.

        :param code: The asset code.
        :raises: :exc:`AssetCodeInvalidError <stellar_sdk.exceptions.AssetCodeInvalidError>`: if ``code`` is invalid.
        """
        asset_code_re = re.compile(r"^[a-zA-Z0-9]{1,12}$")
        if not asset_code_re.match(code):
            raise AssetCodeInvalidError(
                "Asset code is invalid (maximum alphanumeric, 12 characters at max)."
            )

    @property
    def type(self) -> str:
        """Return the type of the asset, can be one of
        following types: `native`, `credit_alphanum4` or `credit_alphanum12`

        :return: The type of the asset.
        """
        return self._type

    @type.setter
    def type(self, v) -> None:
        raise AttributeError("Asset type is immutable.")

    def guess_asset_type(self) -> str:
        """Return the type of the asset, Can be one of
        following types: ``native``, ``credit_alphanum4`` or ``credit_alphanum12``.

        :return: The type of the asset.
        """
        if self.code == "XLM" and self.issuer is None:
            asset_type = "native"
        elif len(self.code) > 4:
            asset_type = "credit_alphanum12"
        else:
            asset_type = "credit_alphanum4"
        return asset_type

    def to_dict(self) -> dict:
        """Generate a dict for this object's attributes.

        :return: A dict representing an :class:`Asset`
        """
        rv: Dict[str, Optional[str]] = {"type": self.type}
        if not self.is_native():
            rv["code"] = self.code
            rv["issuer"] = self.issuer
        return rv

    @classmethod
    def native(cls) -> "Asset":
        """Returns an asset object for the native asset.

        :return: An asset object for the native asset.
        """
        return cls("XLM")

    def is_native(self) -> bool:
        """Return ``Ture`` if the :class:`Asset` object is the native asset.

        :return: ``True`` if the asset object is native, ``False`` otherwise.
        """
        return self.issuer is None

    def to_xdr_object(self) -> stellar_xdr.Asset:
        """Returns the xdr object for this asset.

        :return: XDR Asset object
        """
        asset_xdr = self._to_xdr_object(stellar_xdr.Asset)
        assert isinstance(asset_xdr, stellar_xdr.Asset)
        return asset_xdr

    def to_change_trust_asset_xdr_object(self) -> stellar_xdr.ChangeTrustAsset:
        """Returns the xdr object for this asset.

        :return: XDR ChangeTrustAsset object
        """
        asset_xdr = self._to_xdr_object(stellar_xdr.ChangeTrustAsset)
        assert isinstance(asset_xdr, stellar_xdr.ChangeTrustAsset)
        return asset_xdr

    def to_trust_line_asset_xdr_object(self) -> stellar_xdr.TrustLineAsset:
        """Returns the xdr object for this asset.

        :return: XDR TrustLineAsset object
        """
        asset_xdr = self._to_xdr_object(stellar_xdr.TrustLineAsset)
        assert isinstance(asset_xdr, stellar_xdr.TrustLineAsset)
        return asset_xdr

    def _to_xdr_object(
        self,
        xdr_asset: Union[
            Type[stellar_xdr.Asset],
            Type[stellar_xdr.ChangeTrustAsset],
            Type[stellar_xdr.TrustLineAsset],
        ],
    ) -> Union[
        stellar_xdr.Asset, stellar_xdr.ChangeTrustAsset, stellar_xdr.TrustLineAsset
    ]:
        if self.is_native():
            asset_type = stellar_xdr.AssetType.ASSET_TYPE_NATIVE
            return xdr_asset(type=asset_type)
        else:
            assert self.issuer is not None
            length = len(self.code)
            pad_length = 4 - length if length <= 4 else 12 - length
            asset_code = bytearray(self.code, "ascii") + b"\x00" * pad_length
            issuer = Keypair.from_public_key(self.issuer).xdr_account_id()
        if length <= 4:
            xdr_type = stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM4
            asset_code4 = stellar_xdr.AssetCode4(asset_code)
            alpha_num4 = stellar_xdr.AlphaNum4(asset_code4, issuer)
            return xdr_asset(type=xdr_type, alpha_num4=alpha_num4)
        else:
            xdr_type = stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM12
            asset_code12 = stellar_xdr.AssetCode12(asset_code)
            alpha_num12 = stellar_xdr.AlphaNum12(asset_code12, issuer)
            return xdr_asset(type=xdr_type, alpha_num12=alpha_num12)

    @classmethod
    def from_xdr_object(
        cls,
        xdr_object: Union[
            stellar_xdr.Asset, stellar_xdr.ChangeTrustAsset, stellar_xdr.TrustLineAsset
        ],
    ) -> "Asset":
        """Create a :class:`Asset` from an XDR Asset/ChangeTrustAsset/TrustLineAsset object.

        Please note that this function only supports processing the following types of assets:

        - ASSET_TYPE_NATIVE
        - ASSET_TYPE_CREDIT_ALPHANUM4
        - ASSET_TYPE_CREDIT_ALPHANUM12

        :param xdr_object: The XDR Asset/ChangeTrustAsset/TrustLineAsset object.
        :return: A new :class:`Asset` object from the given XDR object.
        """
        if xdr_object.type == stellar_xdr.AssetType.ASSET_TYPE_NATIVE:
            return Asset.native()
        elif xdr_object.type == stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM4:
            assert xdr_object.alpha_num4 is not None
            assert xdr_object.alpha_num4.issuer.account_id.ed25519 is not None
            issuer = StrKey.encode_ed25519_public_key(
                xdr_object.alpha_num4.issuer.account_id.ed25519.uint256
            )
            code = xdr_object.alpha_num4.asset_code.asset_code4.decode().rstrip("\x00")
        elif xdr_object.type == stellar_xdr.AssetType.ASSET_TYPE_CREDIT_ALPHANUM12:
            assert xdr_object.alpha_num12 is not None
            assert xdr_object.alpha_num12.issuer.account_id.ed25519 is not None
            issuer = StrKey.encode_ed25519_public_key(
                xdr_object.alpha_num12.issuer.account_id.ed25519.uint256
            )
            code = xdr_object.alpha_num12.asset_code.asset_code12.decode().rstrip(
                "\x00"
            )
        else:
            raise ValueError(f"Invalid asset type: {xdr_object.type}")
        return cls(code, issuer)

    def __hash__(self) -> int:
        return hash((self.type, self.code, self.issuer))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.code == other.code and self.issuer == other.issuer

    def __str__(self):
        return f"<Asset [code={self.code}, issuer={self.issuer}, type={self.type}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .ledger_entry_changes import LedgerEntryChanges
from .operation_meta import OperationMeta

__all__ = ["TransactionMetaV1"]


class TransactionMetaV1:
    """
    XDR Source Code::

        struct TransactionMetaV1
        {
            LedgerEntryChanges txChanges; // tx level changes if any
            OperationMeta operations<>;   // meta for each operation
        };
    """

    def __init__(
        self,
        tx_changes: LedgerEntryChanges,
        operations: List[OperationMeta],
    ) -> None:
        _expect_max_length = 4294967295
        if operations and len(operations) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `operations` should be {_expect_max_length}, but got {len(operations)}."
            )
        self.tx_changes = tx_changes
        self.operations = operations

    def pack(self, packer: Packer) -> None:
        self.tx_changes.pack(packer)
        packer.pack_uint(len(self.operations))
        for operations_item in self.operations:
            operations_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionMetaV1:
        tx_changes = LedgerEntryChanges.unpack(unpacker)
        length = unpacker.unpack_uint()
        operations = []
        for _ in range(length):
            operations.append(OperationMeta.unpack(unpacker))
        return cls(
            tx_changes=tx_changes,
            operations=operations,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionMetaV1:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionMetaV1:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.tx_changes,
                self.operations,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.tx_changes == other.tx_changes and self.operations == other.operations
        )

    def __str__(self):
        out = [
            f"tx_changes={self.tx_changes}",
            f"operations={self.operations}",
        ]
        return f"<TransactionMetaV1 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .int64 import Int64

__all__ = ["LedgerKeyOffer"]


class LedgerKeyOffer:
    """
    XDR Source Code::

        struct
            {
                AccountID sellerID;
                int64 offerID;
            }
    """

    def __init__(
        self,
        seller_id: AccountID,
        offer_id: Int64,
    ) -> None:
        self.seller_id = seller_id
        self.offer_id = offer_id

    def pack(self, packer: Packer) -> None:
        self.seller_id.pack(packer)
        self.offer_id.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerKeyOffer:
        seller_id = AccountID.unpack(unpacker)
        offer_id = Int64.unpack(unpacker)
        return cls(
            seller_id=seller_id,
            offer_id=offer_id,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerKeyOffer:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerKeyOffer:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.seller_id,
                self.offer_id,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.seller_id == other.seller_id and self.offer_id == other.offer_id

    def __str__(self):
        out = [
            f"seller_id={self.seller_id}",
            f"offer_id={self.offer_id}",
        ]
        return f"<LedgerKeyOffer [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import String
from .error_code import ErrorCode

__all__ = ["Error"]


class Error:
    """
    XDR Source Code::

        struct Error
        {
            ErrorCode code;
            string msg<100>;
        };
    """

    def __init__(
        self,
        code: ErrorCode,
        msg: bytes,
    ) -> None:
        self.code = code
        self.msg = msg

    def pack(self, packer: Packer) -> None:
        self.code.pack(packer)
        String(self.msg, 100).pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> Error:
        code = ErrorCode.unpack(unpacker)
        msg = String.unpack(unpacker)
        return cls(
            code=code,
            msg=msg,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> Error:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> Error:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.code,
                self.msg,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.code == other.code and self.msg == other.msg

    def __str__(self):
        out = [
            f"code={self.code}",
            f"msg={self.msg}",
        ]
        return f"<Error [{', '.join(out)}]>"



import copy
from typing import Sequence, Union

from xdrlib3 import Packer

from . import xdr as stellar_xdr
from .base_transaction_envelope import BaseTransactionEnvelope
from .decorated_signature import DecoratedSignature
from .exceptions import SignatureExistError
from .keypair import Keypair
from .signer_key import SignerKeyType
from .transaction import Transaction

__all__ = ["TransactionEnvelope"]


class TransactionEnvelope(BaseTransactionEnvelope["TransactionEnvelope"]):
    """The :class:`TransactionEnvelope` object, which represents a transaction
    envelope ready to sign and submit to send over the network.

    When a transaction is ready to be prepared for sending over the network, it
    must be put into a :class:`TransactionEnvelope`, which includes additional
    metadata such as the signers for a given transaction. Ultimately, this
    class handles signing and conversion to and from XDR for usage on Stellar's
    network.

    :param transaction: The transaction that is encapsulated in this envelope.
    :param list signatures: which contains a list of signatures that have
          already been created.
    :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.
    """

    def __init__(
        self,
        transaction: Transaction,
        network_passphrase: str,
        signatures: Sequence[DecoratedSignature] = None,
    ) -> None:
        super().__init__(network_passphrase, signatures)
        self.transaction: Transaction = transaction

    def signature_base(self) -> bytes:
        """Get the signature base of this transaction envelope.

        Return the "signature base" of this transaction, which is the value
        that, when hashed, should be signed to create a signature that
        validators on the Stellar Network will accept.

        It is composed of a 4 prefix bytes followed by the xdr-encoded form of
        this transaction.

        :return: The signature base of this transaction envelope.

        """
        network_id = self._network_id
        tx = self.transaction
        if isinstance(self.transaction, Transaction) and not self.transaction.v1:
            tx = Transaction.from_xdr_object(self.transaction.to_xdr_object(), v1=False)
            tx.v1 = True
        packer = Packer()
        stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX.pack(packer)
        tx.to_xdr_object().pack(packer)
        return network_id + packer.get_buffer()

    def sign_extra_signers_payload(self, signer: Union[Keypair, str]) -> None:
        """Sign this extra signers' payload with a given keypair.

        Note that the signature must not already be in this instance's list of
        signatures.

        :param signer: The keypair or secret to use for signing this extra signers' payload.
        :raise: :exc:`SignatureExistError <stellar_sdk.exception.SignatureExistError>`:
            if this signature already exists.
        """
        if isinstance(signer, str):
            signer = Keypair.from_secret(signer)
        signer_account_id = signer.public_key
        if (
            not self.transaction.preconditions
            or not self.transaction.preconditions.extra_signers
        ):
            return
        for extra_signer in self.transaction.preconditions.extra_signers:
            if (
                extra_signer.signer_key_type
                != SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD
            ):
                continue

            signed_payload_signer_key = extra_signer.to_signed_payload_signer()
            if signed_payload_signer_key.account_id != signer_account_id:
                continue
            sig = signer.sign_payload_decorated(signed_payload_signer_key.payload)
            if sig in self.signatures:
                raise SignatureExistError("The keypair has already signed.")
            else:
                self.signatures.append(sig)

    def to_xdr_object(self) -> stellar_xdr.TransactionEnvelope:
        """Get an XDR object representation of this :class:`TransactionEnvelope`.

        :return: XDR TransactionEnvelope object
        """
        tx = self.transaction.to_xdr_object()
        signatures = [signature.to_xdr_object() for signature in self.signatures]
        if self.transaction.v1:
            assert isinstance(tx, stellar_xdr.Transaction)
            te_type = stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX
            tx_v1_envelope = stellar_xdr.TransactionV1Envelope(tx, signatures)
            return stellar_xdr.TransactionEnvelope(type=te_type, v1=tx_v1_envelope)
        else:
            assert isinstance(tx, stellar_xdr.TransactionV0)
            te_type = stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_V0
            tx_v0_envelope = stellar_xdr.TransactionV0Envelope(tx, signatures)
            return stellar_xdr.TransactionEnvelope(type=te_type, v0=tx_v0_envelope)

    def to_transaction_envelope_v1(self) -> "TransactionEnvelope":
        """Create a new :class:`TransactionEnvelope`, if the internal tx is not v1, we will convert it to v1."""
        tx = copy.deepcopy(self.transaction)
        tx.v1 = True
        return TransactionEnvelope(tx, self.network_passphrase, self.signatures)

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.TransactionEnvelope, network_passphrase: str
    ) -> "TransactionEnvelope":
        """Create a new :class:`TransactionEnvelope` from an XDR object.

        :param xdr_object: The XDR object that represents a transaction envelope.
        :param network_passphrase: The network to connect to for verifying and retrieving additional attributes from.

        :return: A new :class:`TransactionEnvelope` object from the given XDR TransactionEnvelope object.
        """
        te_type = xdr_object.type
        if te_type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX_V0:
            assert xdr_object.v0 is not None
            assert xdr_object.v0.signatures is not None
            tx = Transaction.from_xdr_object(xdr_object.v0.tx, v1=False)
            signatures = [
                DecoratedSignature.from_xdr_object(s) for s in xdr_object.v0.signatures
            ]
        elif te_type == stellar_xdr.EnvelopeType.ENVELOPE_TYPE_TX:
            assert xdr_object.v1 is not None
            assert xdr_object.v1.signatures is not None
            tx = Transaction.from_xdr_object(xdr_object.v1.tx, v1=True)
            signatures = [
                DecoratedSignature.from_xdr_object(s) for s in xdr_object.v1.signatures
            ]
        else:
            raise ValueError(f"Unexpected EnvelopeType: {xdr_object.type}.")
        te = cls(tx, network_passphrase=network_passphrase, signatures=signatures)
        return te

    def __hash__(self):
        return hash((self.transaction, self.network_passphrase, self.signatures))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.transaction == other.transaction
            and self.network_passphrase == other.network_passphrase
            and self.signatures == other.signatures
        )

    def __str__(self):
        return (
            f"<TransactionEnvelope [transaction={self.transaction}, "
            f"network_passphrase={self.network_passphrase}, signatures={self.signatures}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import String
from .node_id import NodeID
from .uint64 import Uint64

__all__ = ["PeerStats"]


class PeerStats:
    """
    XDR Source Code::

        struct PeerStats
        {
            NodeID id;
            string versionStr<100>;
            uint64 messagesRead;
            uint64 messagesWritten;
            uint64 bytesRead;
            uint64 bytesWritten;
            uint64 secondsConnected;

            uint64 uniqueFloodBytesRecv;
            uint64 duplicateFloodBytesRecv;
            uint64 uniqueFetchBytesRecv;
            uint64 duplicateFetchBytesRecv;

            uint64 uniqueFloodMessageRecv;
            uint64 duplicateFloodMessageRecv;
            uint64 uniqueFetchMessageRecv;
            uint64 duplicateFetchMessageRecv;
        };
    """

    def __init__(
        self,
        id: NodeID,
        version_str: bytes,
        messages_read: Uint64,
        messages_written: Uint64,
        bytes_read: Uint64,
        bytes_written: Uint64,
        seconds_connected: Uint64,
        unique_flood_bytes_recv: Uint64,
        duplicate_flood_bytes_recv: Uint64,
        unique_fetch_bytes_recv: Uint64,
        duplicate_fetch_bytes_recv: Uint64,
        unique_flood_message_recv: Uint64,
        duplicate_flood_message_recv: Uint64,
        unique_fetch_message_recv: Uint64,
        duplicate_fetch_message_recv: Uint64,
    ) -> None:
        self.id = id
        self.version_str = version_str
        self.messages_read = messages_read
        self.messages_written = messages_written
        self.bytes_read = bytes_read
        self.bytes_written = bytes_written
        self.seconds_connected = seconds_connected
        self.unique_flood_bytes_recv = unique_flood_bytes_recv
        self.duplicate_flood_bytes_recv = duplicate_flood_bytes_recv
        self.unique_fetch_bytes_recv = unique_fetch_bytes_recv
        self.duplicate_fetch_bytes_recv = duplicate_fetch_bytes_recv
        self.unique_flood_message_recv = unique_flood_message_recv
        self.duplicate_flood_message_recv = duplicate_flood_message_recv
        self.unique_fetch_message_recv = unique_fetch_message_recv
        self.duplicate_fetch_message_recv = duplicate_fetch_message_recv

    def pack(self, packer: Packer) -> None:
        self.id.pack(packer)
        String(self.version_str, 100).pack(packer)
        self.messages_read.pack(packer)
        self.messages_written.pack(packer)
        self.bytes_read.pack(packer)
        self.bytes_written.pack(packer)
        self.seconds_connected.pack(packer)
        self.unique_flood_bytes_recv.pack(packer)
        self.duplicate_flood_bytes_recv.pack(packer)
        self.unique_fetch_bytes_recv.pack(packer)
        self.duplicate_fetch_bytes_recv.pack(packer)
        self.unique_flood_message_recv.pack(packer)
        self.duplicate_flood_message_recv.pack(packer)
        self.unique_fetch_message_recv.pack(packer)
        self.duplicate_fetch_message_recv.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PeerStats:
        id = NodeID.unpack(unpacker)
        version_str = String.unpack(unpacker)
        messages_read = Uint64.unpack(unpacker)
        messages_written = Uint64.unpack(unpacker)
        bytes_read = Uint64.unpack(unpacker)
        bytes_written = Uint64.unpack(unpacker)
        seconds_connected = Uint64.unpack(unpacker)
        unique_flood_bytes_recv = Uint64.unpack(unpacker)
        duplicate_flood_bytes_recv = Uint64.unpack(unpacker)
        unique_fetch_bytes_recv = Uint64.unpack(unpacker)
        duplicate_fetch_bytes_recv = Uint64.unpack(unpacker)
        unique_flood_message_recv = Uint64.unpack(unpacker)
        duplicate_flood_message_recv = Uint64.unpack(unpacker)
        unique_fetch_message_recv = Uint64.unpack(unpacker)
        duplicate_fetch_message_recv = Uint64.unpack(unpacker)
        return cls(
            id=id,
            version_str=version_str,
            messages_read=messages_read,
            messages_written=messages_written,
            bytes_read=bytes_read,
            bytes_written=bytes_written,
            seconds_connected=seconds_connected,
            unique_flood_bytes_recv=unique_flood_bytes_recv,
            duplicate_flood_bytes_recv=duplicate_flood_bytes_recv,
            unique_fetch_bytes_recv=unique_fetch_bytes_recv,
            duplicate_fetch_bytes_recv=duplicate_fetch_bytes_recv,
            unique_flood_message_recv=unique_flood_message_recv,
            duplicate_flood_message_recv=duplicate_flood_message_recv,
            unique_fetch_message_recv=unique_fetch_message_recv,
            duplicate_fetch_message_recv=duplicate_fetch_message_recv,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PeerStats:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PeerStats:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.id,
                self.version_str,
                self.messages_read,
                self.messages_written,
                self.bytes_read,
                self.bytes_written,
                self.seconds_connected,
                self.unique_flood_bytes_recv,
                self.duplicate_flood_bytes_recv,
                self.unique_fetch_bytes_recv,
                self.duplicate_fetch_bytes_recv,
                self.unique_flood_message_recv,
                self.duplicate_flood_message_recv,
                self.unique_fetch_message_recv,
                self.duplicate_fetch_message_recv,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.id == other.id
            and self.version_str == other.version_str
            and self.messages_read == other.messages_read
            and self.messages_written == other.messages_written
            and self.bytes_read == other.bytes_read
            and self.bytes_written == other.bytes_written
            and self.seconds_connected == other.seconds_connected
            and self.unique_flood_bytes_recv == other.unique_flood_bytes_recv
            and self.duplicate_flood_bytes_recv == other.duplicate_flood_bytes_recv
            and self.unique_fetch_bytes_recv == other.unique_fetch_bytes_recv
            and self.duplicate_fetch_bytes_recv == other.duplicate_fetch_bytes_recv
            and self.unique_flood_message_recv == other.unique_flood_message_recv
            and self.duplicate_flood_message_recv == other.duplicate_flood_message_recv
            and self.unique_fetch_message_recv == other.unique_fetch_message_recv
            and self.duplicate_fetch_message_recv == other.duplicate_fetch_message_recv
        )

    def __str__(self):
        out = [
            f"id={self.id}",
            f"version_str={self.version_str}",
            f"messages_read={self.messages_read}",
            f"messages_written={self.messages_written}",
            f"bytes_read={self.bytes_read}",
            f"bytes_written={self.bytes_written}",
            f"seconds_connected={self.seconds_connected}",
            f"unique_flood_bytes_recv={self.unique_flood_bytes_recv}",
            f"duplicate_flood_bytes_recv={self.duplicate_flood_bytes_recv}",
            f"unique_fetch_bytes_recv={self.unique_fetch_bytes_recv}",
            f"duplicate_fetch_bytes_recv={self.duplicate_fetch_bytes_recv}",
            f"unique_flood_message_recv={self.unique_flood_message_recv}",
            f"duplicate_flood_message_recv={self.duplicate_flood_message_recv}",
            f"unique_fetch_message_recv={self.unique_fetch_message_recv}",
            f"duplicate_fetch_message_recv={self.duplicate_fetch_message_recv}",
        ]
        return f"<PeerStats [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .asset import Asset
from .int64 import Int64
from .muxed_account import MuxedAccount

__all__ = ["PathPaymentStrictSendOp"]


class PathPaymentStrictSendOp:
    """
    XDR Source Code::

        struct PathPaymentStrictSendOp
        {
            Asset sendAsset;  // asset we pay with
            int64 sendAmount; // amount of sendAsset to send (excluding fees)

            MuxedAccount destination; // recipient of the payment
            Asset destAsset;          // what they end up with
            int64 destMin;            // the minimum amount of dest asset to
                                      // be received
                                      // The operation will fail if it can't be met

            Asset path<5>; // additional hops it must go through to get there
        };
    """

    def __init__(
        self,
        send_asset: Asset,
        send_amount: Int64,
        destination: MuxedAccount,
        dest_asset: Asset,
        dest_min: Int64,
        path: List[Asset],
    ) -> None:
        _expect_max_length = 5
        if path and len(path) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `path` should be {_expect_max_length}, but got {len(path)}."
            )
        self.send_asset = send_asset
        self.send_amount = send_amount
        self.destination = destination
        self.dest_asset = dest_asset
        self.dest_min = dest_min
        self.path = path

    def pack(self, packer: Packer) -> None:
        self.send_asset.pack(packer)
        self.send_amount.pack(packer)
        self.destination.pack(packer)
        self.dest_asset.pack(packer)
        self.dest_min.pack(packer)
        packer.pack_uint(len(self.path))
        for path_item in self.path:
            path_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> PathPaymentStrictSendOp:
        send_asset = Asset.unpack(unpacker)
        send_amount = Int64.unpack(unpacker)
        destination = MuxedAccount.unpack(unpacker)
        dest_asset = Asset.unpack(unpacker)
        dest_min = Int64.unpack(unpacker)
        length = unpacker.unpack_uint()
        path = []
        for _ in range(length):
            path.append(Asset.unpack(unpacker))
        return cls(
            send_asset=send_asset,
            send_amount=send_amount,
            destination=destination,
            dest_asset=dest_asset,
            dest_min=dest_min,
            path=path,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> PathPaymentStrictSendOp:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> PathPaymentStrictSendOp:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.send_asset,
                self.send_amount,
                self.destination,
                self.dest_asset,
                self.dest_min,
                self.path,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.send_asset == other.send_asset
            and self.send_amount == other.send_amount
            and self.destination == other.destination
            and self.dest_asset == other.dest_asset
            and self.dest_min == other.dest_min
            and self.path == other.path
        )

    def __str__(self):
        out = [
            f"send_asset={self.send_asset}",
            f"send_amount={self.send_amount}",
            f"destination={self.destination}",
            f"dest_asset={self.dest_asset}",
            f"dest_min={self.dest_min}",
            f"path={self.path}",
        ]
        return f"<PathPaymentStrictSendOp [{', '.join(out)}]>"



from ...asset import Asset
from ...call_builder.base.base_call_builder import BaseCallBuilder
from ...utils import convert_assets_to_horizon_param

__all__ = ["BaseAccountsCallBuilder"]


class BaseAccountsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`AccountsCallBuilder` pointed to server defined by horizon_url.

    See `List All Accounts <https://developers.stellar.org/api/resources/accounts/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "accounts"

    def account_id(self, account_id: str):
        """Returns information and links relating to a single account.
        The balances section in the returned JSON will also list all the trust lines this account has set up.

        See `Retrieve an Account <https://developers.stellar.org/api/resources/accounts/single/>`__ for more information.

        :param account_id: account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: current AccountCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}"
        return self

    def for_signer(self, signer: str):
        """Filtering accounts who have a given signer. The result is a list of accounts.

        See `List All Accounts <https://developers.stellar.org/api/resources/accounts/list/>`__ for more information.

        :param signer: signer's account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: current AccountCallBuilder instance
        """
        self._add_query_param("signer", signer)
        return self

    def for_asset(self, asset: Asset):
        """Filtering accounts who have a trustline to an asset. The result is a list of accounts.

        See `List All Accounts <https://developers.stellar.org/api/resources/accounts/list/>`__ for more information.

        :param asset: an issued asset
        :return: current AccountCallBuilder instance
        """
        assets_param = convert_assets_to_horizon_param([asset])
        self._add_query_param("asset", assets_param)
        return self

    def for_sponsor(self, sponsor: str):
        """Filtering accounts where the given account is sponsoring the account or any of its sub-entries.

        See `List All Accounts <https://developers.stellar.org/api/resources/accounts/list/>`__ for more information.

        :param sponsor: the sponsor id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: current AccountCallBuilder instance
        """
        self._add_query_param("sponsor", sponsor)
        return self

    def for_liquidity_pool(self, liquidity_pool_id: str):
        """Filtering accounts who have a trustline for the given pool. The result is a list of accounts.

        See `List All Accounts <https://developers.stellar.org/api/resources/accounts/list/>`__ for more information.

        :param liquidity_pool_id: The ID of the liquidity pool in hex string.,
            for example: ``"dd7b1ab831c273310ddbec6f97870aa83c2fbd78ce22aded37ecbf4f3380fac7"``
        :return: current AccountCallBuilder instance
        """
        self._add_query_param("liquidity_pool", liquidity_pool_id)
        return self



from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["Inflation"]


class Inflation(Operation):
    """The :class:`Inflation` object, which represents a
    Inflation operation on Stellar's network.

    This operation runs inflation.

    Threshold: Low

    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = stellar_xdr.OperationType.INFLATION

    def __init__(self, source: Optional[Union[MuxedAccount, str]] = None) -> None:
        super().__init__(source)

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        body = stellar_xdr.OperationBody(type=self._XDR_OPERATION_TYPE)
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "Inflation":
        """Creates a :class:`Inflation` object from an XDR Operation
        object.

        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        op = cls(source)
        return op

    def __str__(self):
        return f"<Inflation [source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .account_id import AccountID
from .int64 import Int64

__all__ = ["InflationPayout"]


class InflationPayout:
    """
    XDR Source Code::

        struct InflationPayout // or use PaymentResultAtom to limit types?
        {
            AccountID destination;
            int64 amount;
        };
    """

    def __init__(
        self,
        destination: AccountID,
        amount: Int64,
    ) -> None:
        self.destination = destination
        self.amount = amount

    def pack(self, packer: Packer) -> None:
        self.destination.pack(packer)
        self.amount.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> InflationPayout:
        destination = AccountID.unpack(unpacker)
        amount = Int64.unpack(unpacker)
        return cls(
            destination=destination,
            amount=amount,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> InflationPayout:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> InflationPayout:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.destination,
                self.amount,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.destination == other.destination and self.amount == other.amount

    def __str__(self):
        out = [
            f"destination={self.destination}",
            f"amount={self.amount}",
        ]
        return f"<InflationPayout [{', '.join(out)}]>"



from ...asset import Asset
from ...call_builder.base.base_call_builder import BaseCallBuilder
from ...utils import convert_assets_to_horizon_param

__all__ = ["BaseClaimableBalancesCallBuilder"]


class BaseClaimableBalancesCallBuilder(BaseCallBuilder):
    """Creates a new :class:`ClaimableBalancesCallBuilder` pointed to server defined by horizon_url.

    See `List Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "claimable_balances"

    def claimable_balance(self, claimable_balance_id: str):
        """Returns information and links relating to a single claimable balance.

        See `List Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`__ for more information.

        :param claimable_balance_id: claimable balance id
        :return: current AccountCallBuilder instance
        """
        self.endpoint = f"claimable_balances/{claimable_balance_id}"
        return self

    def for_sponsor(self, sponsor: str):
        """Returns all claimable balances which are sponsored by the given account ID.

        See `List Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`__ for more information.

        :param sponsor: the sponsor id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: current ClaimableBalancesCallBuilder instance
        """
        self._add_query_param("sponsor", sponsor)
        return self

    def for_asset(self, asset: Asset):
        """Returns all claimable balances which provide a balance for the given asset.

        See `List Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`__ for more information.

        :param asset: an asset
        :return: current ClaimableBalancesCallBuilder instance
        """
        assets_param = convert_assets_to_horizon_param([asset])
        self._add_query_param("asset", assets_param)
        return self

    def for_claimant(self, claimant: str):
        """Returns all claimable balances which can be claimed by the given account ID.

        See `List Claimable Balances <https://developers.stellar.org/api/resources/claimablebalances/list/>`__ for more information.

        :param claimant: the account id, for example: ``"GDGQVOKHW4VEJRU2TETD6DBRKEO5ERCNF353LW5WBFW3JJWQ2BRQ6KDD"``
        :return: current ClaimableBalancesCallBuilder instance
        """
        self._add_query_param("claimant", claimant)
        return self



from decimal import Decimal
from typing import Union

from . import xdr as stellar_xdr
from .utils import best_rational_approximation

__all__ = ["Price"]


class Price:
    """Create a new price. Price in Stellar is represented as a fraction.

    An example::

        from stellar_sdk import Price

        price_a = Price(1, 2)
        price_b = Price.from_raw_price("0.5")

    :param n: numerator
    :param d: denominator
    """

    def __init__(self, n: int, d: int) -> None:
        self.n: int = n
        self.d: int = d

    @classmethod
    def from_raw_price(cls, price: Union[str, Decimal]) -> "Price":
        """Create a :class:`Price` from the given str or Decimal price.

        :param price: the str or Decimal price. (ex. ``"0.125"``)
        :return: A new :class:`Price` object from the given str or Decimal price.
        :raises: :exc:`NoApproximationError <stellar_sdk.exceptions.NoApproximationError>`:
            if the approximation could not not be found.
        """
        best_r = best_rational_approximation(str(price))
        n = best_r["n"]
        d = best_r["d"]
        return cls(n, d)

    def to_xdr_object(self) -> stellar_xdr.Price:
        """Returns the xdr object for this price object.

        :return: XDR Price object
        """
        return stellar_xdr.Price(
            n=stellar_xdr.Int32(self.n), d=stellar_xdr.Int32(self.d)
        )

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Price) -> "Price":
        """Create a :class:`Price` from an XDR Price object.

        :param xdr_object: The XDR Price object.
        :return: A new :class:`Price` object from the given XDR Price object.
        """
        n = xdr_object.n.int32
        d = xdr_object.d.int32
        return cls(n, d)

    def __lt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.n * other.d) < (other.n * self.d)

    def __le__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.n * other.d) <= (other.n * self.d)

    def __hash__(self):
        return hash((self.n, self.d))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.n * other.d) == (other.n * self.d)

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.n * other.d) != (other.n * self.d)

    def __gt__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.n * other.d) > (other.n * self.d)

    def __ge__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (self.n * other.d) >= (other.n * self.d)

    def __str__(self):
        return f"<Price [n={self.n}, d={self.d}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import Optional

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .scp_ballot import SCPBallot
from .uint32 import Uint32

__all__ = ["SCPStatementPrepare"]


class SCPStatementPrepare:
    """
    XDR Source Code::

        struct
                {
                    Hash quorumSetHash;       // D
                    SCPBallot ballot;         // b
                    SCPBallot* prepared;      // p
                    SCPBallot* preparedPrime; // p'
                    uint32 nC;                // c.n
                    uint32 nH;                // h.n
                }
    """

    def __init__(
        self,
        quorum_set_hash: Hash,
        ballot: SCPBallot,
        prepared: Optional[SCPBallot],
        prepared_prime: Optional[SCPBallot],
        n_c: Uint32,
        n_h: Uint32,
    ) -> None:
        self.quorum_set_hash = quorum_set_hash
        self.ballot = ballot
        self.prepared = prepared
        self.prepared_prime = prepared_prime
        self.n_c = n_c
        self.n_h = n_h

    def pack(self, packer: Packer) -> None:
        self.quorum_set_hash.pack(packer)
        self.ballot.pack(packer)
        if self.prepared is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.prepared.pack(packer)
        if self.prepared_prime is None:
            packer.pack_uint(0)
        else:
            packer.pack_uint(1)
            self.prepared_prime.pack(packer)
        self.n_c.pack(packer)
        self.n_h.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCPStatementPrepare:
        quorum_set_hash = Hash.unpack(unpacker)
        ballot = SCPBallot.unpack(unpacker)
        prepared = SCPBallot.unpack(unpacker) if unpacker.unpack_uint() else None
        prepared_prime = SCPBallot.unpack(unpacker) if unpacker.unpack_uint() else None
        n_c = Uint32.unpack(unpacker)
        n_h = Uint32.unpack(unpacker)
        return cls(
            quorum_set_hash=quorum_set_hash,
            ballot=ballot,
            prepared=prepared,
            prepared_prime=prepared_prime,
            n_c=n_c,
            n_h=n_h,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCPStatementPrepare:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCPStatementPrepare:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.quorum_set_hash,
                self.ballot,
                self.prepared,
                self.prepared_prime,
                self.n_c,
                self.n_h,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.quorum_set_hash == other.quorum_set_hash
            and self.ballot == other.ballot
            and self.prepared == other.prepared
            and self.prepared_prime == other.prepared_prime
            and self.n_c == other.n_c
            and self.n_h == other.n_h
        )

    def __str__(self):
        out = [
            f"quorum_set_hash={self.quorum_set_hash}",
            f"ballot={self.ballot}",
            f"prepared={self.prepared}",
            f"prepared_prime={self.prepared_prime}",
            f"n_c={self.n_c}",
            f"n_h={self.n_h}",
        ]
        return f"<SCPStatementPrepare [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .sc_map_entry import SCMapEntry

__all__ = ["SCMap"]


class SCMap:
    """
    XDR Source Code::

        typedef SCMapEntry SCMap<>;
    """

    def __init__(self, sc_map: List[SCMapEntry]) -> None:
        _expect_max_length = 4294967295
        if sc_map and len(sc_map) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `sc_map` should be {_expect_max_length}, but got {len(sc_map)}."
            )
        self.sc_map = sc_map

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.sc_map))
        for sc_map_item in self.sc_map:
            sc_map_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCMap:
        length = unpacker.unpack_uint()
        sc_map = []
        for _ in range(length):
            sc_map.append(SCMapEntry.unpack(unpacker))
        return cls(sc_map)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCMap:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCMap:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(self.sc_map)

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.sc_map == other.sc_map

    def __str__(self):
        return f"<SCMap [sc_map={self.sc_map}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .ledger_entry_extension_v1_ext import LedgerEntryExtensionV1Ext
from .sponsorship_descriptor import SponsorshipDescriptor

__all__ = ["LedgerEntryExtensionV1"]


class LedgerEntryExtensionV1:
    """
    XDR Source Code::

        struct LedgerEntryExtensionV1
        {
            SponsorshipDescriptor sponsoringID;

            union switch (int v)
            {
            case 0:
                void;
            }
            ext;
        };
    """

    def __init__(
        self,
        sponsoring_id: SponsorshipDescriptor,
        ext: LedgerEntryExtensionV1Ext,
    ) -> None:
        self.sponsoring_id = sponsoring_id
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.sponsoring_id.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerEntryExtensionV1:
        sponsoring_id = SponsorshipDescriptor.unpack(unpacker)
        ext = LedgerEntryExtensionV1Ext.unpack(unpacker)
        return cls(
            sponsoring_id=sponsoring_id,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerEntryExtensionV1:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerEntryExtensionV1:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.sponsoring_id,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.sponsoring_id == other.sponsoring_id and self.ext == other.ext

    def __str__(self):
        out = [
            f"sponsoring_id={self.sponsoring_id}",
            f"ext={self.ext}",
        ]
        return f"<LedgerEntryExtensionV1 [{', '.join(out)}]>"



import binascii
from decimal import Decimal
from typing import Optional, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from ..price import Price
from ..utils import raise_if_not_valid_amount, raise_if_not_valid_hash
from .operation import Operation

__all__ = ["LiquidityPoolDeposit"]


class LiquidityPoolDeposit(Operation):
    """The :class:`LiquidityPoolDeposit` object, which represents a LiquidityPoolDeposit
    operation on Stellar's network.

    Creates a liquidity pool deposit operation.

    Threshold: Medium

    See `Liquidity Pool Deposit <https://developers.stellar.org/docs/start/list-of-operations/#liquidity-pool-deposit>`_ for more information.

    :param liquidity_pool_id: The liquidity pool ID.
    :param max_amount_a: Maximum amount of first asset to deposit.
    :param max_amount_b: Maximum amount of second asset to deposit.
    :param min_price: Minimum deposit_a/deposit_b price.
    :param max_price: Maximum deposit_a/deposit_b price.
    :param source: The source account for the operation. Defaults to the
        transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.LIQUIDITY_POOL_DEPOSIT
    )

    def __init__(
        self,
        liquidity_pool_id: str,
        max_amount_a: Union[str, Decimal],
        max_amount_b: Union[str, Decimal],
        min_price: Union[str, Decimal, Price],
        max_price: Union[str, Decimal, Price],
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        super().__init__(source)
        self.liquidity_pool_id: str = liquidity_pool_id
        self.max_amount_a: str = str(max_amount_a)
        self.max_amount_b: str = str(max_amount_b)
        if isinstance(min_price, Price):
            self.min_price: Price = min_price
        else:
            self.min_price = Price.from_raw_price(min_price)
        if isinstance(max_price, Price):
            self.max_price: Price = max_price
        else:
            self.max_price = Price.from_raw_price(max_price)
        raise_if_not_valid_amount(self.max_amount_a, "max_amount_a")
        raise_if_not_valid_amount(self.max_amount_b, "max_amount_b")
        raise_if_not_valid_hash(self.liquidity_pool_id, "liquidity_pool_id")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        liquidity_pool_id_bytes = binascii.unhexlify(self.liquidity_pool_id)
        liquidity_pool_id = stellar_xdr.PoolID.from_xdr_bytes(liquidity_pool_id_bytes)
        max_amount_a = stellar_xdr.Int64(Operation.to_xdr_amount(self.max_amount_a))
        max_amount_b = stellar_xdr.Int64(Operation.to_xdr_amount(self.max_amount_b))
        min_price = self.min_price.to_xdr_object()
        max_price = self.max_price.to_xdr_object()
        liquidity_pool_deposit_op = stellar_xdr.LiquidityPoolDepositOp(
            liquidity_pool_id, max_amount_a, max_amount_b, min_price, max_price
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            liquidity_pool_deposit_op=liquidity_pool_deposit_op,
        )
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "LiquidityPoolDeposit":
        """Creates a :class:`LiquidityPoolDeposit` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.liquidity_pool_deposit_op is not None
        liquidity_pool_id_bytes = (
            xdr_object.body.liquidity_pool_deposit_op.liquidity_pool_id.pool_id.hash
        )
        liquidity_pool_id = liquidity_pool_id_bytes.hex()
        max_amount_a = Operation.from_xdr_amount(
            xdr_object.body.liquidity_pool_deposit_op.max_amount_a.int64
        )
        max_amount_b = Operation.from_xdr_amount(
            xdr_object.body.liquidity_pool_deposit_op.max_amount_b.int64
        )
        min_price = Price.from_xdr_object(
            xdr_object.body.liquidity_pool_deposit_op.min_price
        )
        max_price = Price.from_xdr_object(
            xdr_object.body.liquidity_pool_deposit_op.max_price
        )

        op = cls(
            source=source,
            liquidity_pool_id=liquidity_pool_id,
            max_amount_a=max_amount_a,
            max_amount_b=max_amount_b,
            min_price=min_price,
            max_price=max_price,
        )
        return op

    def __str__(self):
        return (
            f"<LiquidityPoolDeposit [liquidity_pool_id={self.liquidity_pool_id}, max_amount_a={self.max_amount_a}, "
            f"max_amount_b={self.max_amount_b}, min_price={self.min_price}, max_price={self.max_price}, "
            f"source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .extension_point import ExtensionPoint
from .time_point import TimePoint
from .uint32 import Uint32

__all__ = ["AccountEntryExtensionV3"]


class AccountEntryExtensionV3:
    """
    XDR Source Code::

        struct AccountEntryExtensionV3
        {
            // We can use this to add more fields, or because it is first, to
            // change AccountEntryExtensionV3 into a union.
            ExtensionPoint ext;

            // Ledger number at which `seqNum` took on its present value.
            uint32 seqLedger;

            // Time at which `seqNum` took on its present value.
            TimePoint seqTime;
        };
    """

    def __init__(
        self,
        ext: ExtensionPoint,
        seq_ledger: Uint32,
        seq_time: TimePoint,
    ) -> None:
        self.ext = ext
        self.seq_ledger = seq_ledger
        self.seq_time = seq_time

    def pack(self, packer: Packer) -> None:
        self.ext.pack(packer)
        self.seq_ledger.pack(packer)
        self.seq_time.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AccountEntryExtensionV3:
        ext = ExtensionPoint.unpack(unpacker)
        seq_ledger = Uint32.unpack(unpacker)
        seq_time = TimePoint.unpack(unpacker)
        return cls(
            ext=ext,
            seq_ledger=seq_ledger,
            seq_time=seq_time,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AccountEntryExtensionV3:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AccountEntryExtensionV3:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ext,
                self.seq_ledger,
                self.seq_time,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ext == other.ext
            and self.seq_ledger == other.seq_ledger
            and self.seq_time == other.seq_time
        )

    def __str__(self):
        out = [
            f"ext={self.ext}",
            f"seq_ledger={self.seq_ledger}",
            f"seq_time={self.seq_time}",
        ]
        return f"<AccountEntryExtensionV3 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .int64 import Int64
from .soroban_authorized_invocation import SorobanAuthorizedInvocation
from .uint32 import Uint32

__all__ = ["HashIDPreimageSorobanAuthorization"]


class HashIDPreimageSorobanAuthorization:
    """
    XDR Source Code::

        struct
            {
                Hash networkID;
                int64 nonce;
                uint32 signatureExpirationLedger;
                SorobanAuthorizedInvocation invocation;
            }
    """

    def __init__(
        self,
        network_id: Hash,
        nonce: Int64,
        signature_expiration_ledger: Uint32,
        invocation: SorobanAuthorizedInvocation,
    ) -> None:
        self.network_id = network_id
        self.nonce = nonce
        self.signature_expiration_ledger = signature_expiration_ledger
        self.invocation = invocation

    def pack(self, packer: Packer) -> None:
        self.network_id.pack(packer)
        self.nonce.pack(packer)
        self.signature_expiration_ledger.pack(packer)
        self.invocation.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> HashIDPreimageSorobanAuthorization:
        network_id = Hash.unpack(unpacker)
        nonce = Int64.unpack(unpacker)
        signature_expiration_ledger = Uint32.unpack(unpacker)
        invocation = SorobanAuthorizedInvocation.unpack(unpacker)
        return cls(
            network_id=network_id,
            nonce=nonce,
            signature_expiration_ledger=signature_expiration_ledger,
            invocation=invocation,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> HashIDPreimageSorobanAuthorization:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> HashIDPreimageSorobanAuthorization:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.network_id,
                self.nonce,
                self.signature_expiration_ledger,
                self.invocation,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.network_id == other.network_id
            and self.nonce == other.nonce
            and self.signature_expiration_ledger == other.signature_expiration_ledger
            and self.invocation == other.invocation
        )

    def __str__(self):
        out = [
            f"network_id={self.network_id}",
            f"nonce={self.nonce}",
            f"signature_expiration_ledger={self.signature_expiration_ledger}",
            f"invocation={self.invocation}",
        ]
        return f"<HashIDPreimageSorobanAuthorization [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Integer
from .soroban_transaction_data import SorobanTransactionData

__all__ = ["TransactionExt"]


class TransactionExt:
    """
    XDR Source Code::

        union switch (int v)
            {
            case 0:
                void;
            case 1:
                SorobanTransactionData sorobanData;
            }
    """

    def __init__(
        self,
        v: int,
        soroban_data: SorobanTransactionData = None,
    ) -> None:
        self.v = v
        self.soroban_data = soroban_data

    def pack(self, packer: Packer) -> None:
        Integer(self.v).pack(packer)
        if self.v == 0:
            return
        if self.v == 1:
            if self.soroban_data is None:
                raise ValueError("soroban_data should not be None.")
            self.soroban_data.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionExt:
        v = Integer.unpack(unpacker)
        if v == 0:
            return cls(v=v)
        if v == 1:
            soroban_data = SorobanTransactionData.unpack(unpacker)
            return cls(v=v, soroban_data=soroban_data)
        return cls(v=v)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionExt:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionExt:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.v,
                self.soroban_data,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.v == other.v and self.soroban_data == other.soroban_data

    def __str__(self):
        out = []
        out.append(f"v={self.v}")
        out.append(
            f"soroban_data={self.soroban_data}"
        ) if self.soroban_data is not None else None
        return f"<TransactionExt [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List, Optional

from xdrlib3 import Packer, Unpacker

from .claim_predicate_type import ClaimPredicateType
from .int64 import Int64

__all__ = ["ClaimPredicate"]


class ClaimPredicate:
    """
    XDR Source Code::

        union ClaimPredicate switch (ClaimPredicateType type)
        {
        case CLAIM_PREDICATE_UNCONDITIONAL:
            void;
        case CLAIM_PREDICATE_AND:
            ClaimPredicate andPredicates<2>;
        case CLAIM_PREDICATE_OR:
            ClaimPredicate orPredicates<2>;
        case CLAIM_PREDICATE_NOT:
            ClaimPredicate* notPredicate;
        case CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME:
            int64 absBefore; // Predicate will be true if closeTime < absBefore
        case CLAIM_PREDICATE_BEFORE_RELATIVE_TIME:
            int64 relBefore; // Seconds since closeTime of the ledger in which the
                             // ClaimableBalanceEntry was created
        };
    """

    def __init__(
        self,
        type: ClaimPredicateType,
        and_predicates: List["ClaimPredicate"] = None,
        or_predicates: List["ClaimPredicate"] = None,
        not_predicate: Optional["ClaimPredicate"] = None,
        abs_before: Int64 = None,
        rel_before: Int64 = None,
    ) -> None:
        _expect_max_length = 2
        if and_predicates and len(and_predicates) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `and_predicates` should be {_expect_max_length}, but got {len(and_predicates)}."
            )
        _expect_max_length = 2
        if or_predicates and len(or_predicates) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `or_predicates` should be {_expect_max_length}, but got {len(or_predicates)}."
            )
        self.type = type
        self.and_predicates = and_predicates
        self.or_predicates = or_predicates
        self.not_predicate = not_predicate
        self.abs_before = abs_before
        self.rel_before = rel_before

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL:
            return
        if self.type == ClaimPredicateType.CLAIM_PREDICATE_AND:
            if self.and_predicates is None:
                raise ValueError("and_predicates should not be None.")
            packer.pack_uint(len(self.and_predicates))
            for and_predicates_item in self.and_predicates:
                and_predicates_item.pack(packer)
            return
        if self.type == ClaimPredicateType.CLAIM_PREDICATE_OR:
            if self.or_predicates is None:
                raise ValueError("or_predicates should not be None.")
            packer.pack_uint(len(self.or_predicates))
            for or_predicates_item in self.or_predicates:
                or_predicates_item.pack(packer)
            return
        if self.type == ClaimPredicateType.CLAIM_PREDICATE_NOT:
            if self.not_predicate is None:
                packer.pack_uint(0)
            else:
                packer.pack_uint(1)
                if self.not_predicate is None:
                    raise ValueError("not_predicate should not be None.")
                self.not_predicate.pack(packer)
            return
        if self.type == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME:
            if self.abs_before is None:
                raise ValueError("abs_before should not be None.")
            self.abs_before.pack(packer)
            return
        if self.type == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME:
            if self.rel_before is None:
                raise ValueError("rel_before should not be None.")
            self.rel_before.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ClaimPredicate:
        type = ClaimPredicateType.unpack(unpacker)
        if type == ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL:
            return cls(type=type)
        if type == ClaimPredicateType.CLAIM_PREDICATE_AND:
            length = unpacker.unpack_uint()
            and_predicates = []
            for _ in range(length):
                and_predicates.append(ClaimPredicate.unpack(unpacker))
            return cls(type=type, and_predicates=and_predicates)
        if type == ClaimPredicateType.CLAIM_PREDICATE_OR:
            length = unpacker.unpack_uint()
            or_predicates = []
            for _ in range(length):
                or_predicates.append(ClaimPredicate.unpack(unpacker))
            return cls(type=type, or_predicates=or_predicates)
        if type == ClaimPredicateType.CLAIM_PREDICATE_NOT:
            not_predicate = (
                ClaimPredicate.unpack(unpacker) if unpacker.unpack_uint() else None
            )
            return cls(type=type, not_predicate=not_predicate)
        if type == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME:
            abs_before = Int64.unpack(unpacker)
            return cls(type=type, abs_before=abs_before)
        if type == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME:
            rel_before = Int64.unpack(unpacker)
            return cls(type=type, rel_before=rel_before)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ClaimPredicate:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ClaimPredicate:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.and_predicates,
                self.or_predicates,
                self.not_predicate,
                self.abs_before,
                self.rel_before,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.and_predicates == other.and_predicates
            and self.or_predicates == other.or_predicates
            and self.not_predicate == other.not_predicate
            and self.abs_before == other.abs_before
            and self.rel_before == other.rel_before
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(
            f"and_predicates={self.and_predicates}"
        ) if self.and_predicates is not None else None
        out.append(
            f"or_predicates={self.or_predicates}"
        ) if self.or_predicates is not None else None
        out.append(
            f"not_predicate={self.not_predicate}"
        ) if self.not_predicate is not None else None
        out.append(
            f"abs_before={self.abs_before}"
        ) if self.abs_before is not None else None
        out.append(
            f"rel_before={self.rel_before}"
        ) if self.rel_before is not None else None
        return f"<ClaimPredicate [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from enum import IntEnum

from xdrlib3 import Packer, Unpacker

__all__ = ["ChangeTrustResultCode"]


class ChangeTrustResultCode(IntEnum):
    """
    XDR Source Code::

        enum ChangeTrustResultCode
        {
            // codes considered as "success" for the operation
            CHANGE_TRUST_SUCCESS = 0,
            // codes considered as "failure" for the operation
            CHANGE_TRUST_MALFORMED = -1,     // bad input
            CHANGE_TRUST_NO_ISSUER = -2,     // could not find issuer
            CHANGE_TRUST_INVALID_LIMIT = -3, // cannot drop limit below balance
                                             // cannot create with a limit of 0
            CHANGE_TRUST_LOW_RESERVE =
                -4, // not enough funds to create a new trust line,
            CHANGE_TRUST_SELF_NOT_ALLOWED = -5,   // trusting self is not allowed
            CHANGE_TRUST_TRUST_LINE_MISSING = -6, // Asset trustline is missing for pool
            CHANGE_TRUST_CANNOT_DELETE =
                -7, // Asset trustline is still referenced in a pool
            CHANGE_TRUST_NOT_AUTH_MAINTAIN_LIABILITIES =
                -8 // Asset trustline is deauthorized
        };
    """

    CHANGE_TRUST_SUCCESS = 0
    CHANGE_TRUST_MALFORMED = -1
    CHANGE_TRUST_NO_ISSUER = -2
    CHANGE_TRUST_INVALID_LIMIT = -3
    CHANGE_TRUST_LOW_RESERVE = -4
    CHANGE_TRUST_SELF_NOT_ALLOWED = -5
    CHANGE_TRUST_TRUST_LINE_MISSING = -6
    CHANGE_TRUST_CANNOT_DELETE = -7
    CHANGE_TRUST_NOT_AUTH_MAINTAIN_LIABILITIES = -8

    def pack(self, packer: Packer) -> None:
        packer.pack_int(self.value)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ChangeTrustResultCode:
        value = unpacker.unpack_int()
        return cls(value)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ChangeTrustResultCode:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ChangeTrustResultCode:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)



from typing import Union

from ...call_builder.base.base_call_builder import BaseCallBuilder

__all__ = ["BaseOperationsCallBuilder"]


class BaseOperationsCallBuilder(BaseCallBuilder):
    """Creates a new :class:`OperationsCallBuilder` pointed to server defined by horizon_url.

    See `List All Operations <https://developers.stellar.org/api/resources/operations/list/>`__ for more information.

    :param horizon_url: Horizon server URL.
    """

    def __init__(self, horizon_url) -> None:
        super().__init__(horizon_url)
        self.endpoint: str = "operations"

    def operation(self, operation_id: Union[int, str]):
        """The operation details endpoint provides information on a single operation. The operation ID provided
        in the id argument specifies which operation to load.

        See `Retrieve an Operation <https://developers.stellar.org/api/resources/operations/single/>`__ for more information.

        :param operation_id: Operation ID
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"operations/{operation_id}"
        return self

    def for_account(self, account_id: str):
        """This endpoint represents all operations that were included in valid transactions that
        affected a particular account.

        See `Retrieve an Account's Operations <https://developers.stellar.org/api/resources/accounts/operations/>`__ for more information.

        :param account_id: Account ID
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"accounts/{account_id}/operations"
        return self

    def for_ledger(self, sequence: Union[int, str]):
        """This endpoint returns all operations that occurred in a given ledger.

        See `Retrieve a Ledger's Operations <https://developers.stellar.org/api/resources/ledgers/operations/>`__ for more information.

        :param sequence: Sequence ID
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"ledgers/{sequence}/operations"
        return self

    def for_transaction(self, transaction_hash: str):
        """This endpoint represents all operations that are part of a given transaction.

        See `Retrieve a Transaction's Operations <https://developers.stellar.org/api/resources/transactions/operations/>`__ for more information.

        :param transaction_hash: Transaction Hash
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"transactions/{transaction_hash}/operations"
        return self

    def for_claimable_balance(self, claimable_balance_id: str):
        """This endpoint represents successful operations referencing a given
        claimable balance and can be used in streaming mode.


        See `Claimable Balances - Retrieve related Operations <https://developers.stellar.org/api/resources/claimablebalances/operations/>`__ for more information.

        :param claimable_balance_id: This claimable balance’s id encoded in a hex string representation.
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"claimable_balances/{claimable_balance_id}/operations"
        return self

    def for_liquidity_pool(self, liquidity_pool_id: str):
        """This endpoint represents all operations that are part of a given liquidity pool.

        See `Liquidity Pools - Retrieve related Operations <https://developers.stellar.org/api/resources/liquiditypools/operations/>`__ for more information.

        :param liquidity_pool_id: The ID of the liquidity pool in hex string.
        :return: this OperationCallBuilder instance
        """
        self.endpoint = f"liquidity_pools/{liquidity_pool_id}/operations"
        return self

    def include_failed(self, include_failed: bool):
        """Adds a parameter defining whether to include failed transactions. By default only
        operations of successful transactions are returned.

        :param include_failed: Set to `True` to include operations of failed transactions.
        :return: current OperationsCallBuilder instance
        """
        self._add_query_param("include_failed", include_failed)
        return self

    def join(self, join: str):
        """join represents `join` param in queries, currently only supports `transactions`

        :param join: join represents `join` param in queries, currently only supports `transactions`
        :return: current OperationsCallBuilder instance
        """
        self._add_query_param("join", join)
        return self



from enum import IntEnum
from typing import Union

from . import xdr as stellar_xdr
from .__version__ import __issues__
from .strkey import StrKey, _get_version_byte_for_prefix, _VersionByte

__all__ = ["SignerKey", "SignerKeyType", "SignedPayloadSigner"]


class SignerKeyType(IntEnum):
    SIGNER_KEY_TYPE_ED25519 = 0
    SIGNER_KEY_TYPE_PRE_AUTH_TX = 1
    SIGNER_KEY_TYPE_HASH_X = 2
    SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD = 3


class SignedPayloadSigner:
    """The :class:`SignedPayloadSigner` object, which represents a signed payload signer.

    :param account_id: The account id.
    :param payload: The raw payload.
    """

    def __init__(self, account_id: str, payload: bytes) -> None:
        self.account_id: str = account_id
        self.payload: bytes = payload

    def __hash__(self):
        return hash((self.account_id, self.payload))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.account_id == other.account_id and self.payload == self.payload

    def __str__(self):
        return f"<SignedPayloadSigner [account_id={self.account_id}, payload={self.payload}]>"


class SignerKey:
    """The :class:`SignerKey` object, which represents an account signer key on Stellar's network.

    :param signer_key: The signer key.
    :param signer_key: The signer key type.
    """

    def __init__(self, signer_key: bytes, signer_key_type: SignerKeyType) -> None:
        self.signer_key: bytes = signer_key
        self.signer_key_type: SignerKeyType = signer_key_type

    @property
    def encoded_signer_key(self) -> str:
        """
        return: The signer key encoded in Strkey format.
        """
        if self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_ED25519:
            return StrKey.encode_ed25519_public_key(self.signer_key)
        elif self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_PRE_AUTH_TX:
            return StrKey.encode_pre_auth_tx(self.signer_key)
        elif self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_HASH_X:
            return StrKey.encode_sha256_hash(self.signer_key)
        elif (
            self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD
        ):
            return StrKey.encode_ed25519_signed_payload(self.signer_key)
        else:
            raise ValueError(
                f"{self.signer_key_type!r} is an unsupported signer key type."
            )

    @classmethod
    def from_encoded_signer_key(cls, encoded_signer_key: str) -> "SignerKey":
        """Parse the encoded signer key.

        :param encoded_signer_key: The encoded signer key. (ex. ``GBJCHUKZMTFSLOMNC7P4TS4VJJBTCYL3XKSOLXAUJSD56C4LHND5TWUC``)
        :return: The :class:`SignerKey` object.
        """
        prefix = _get_version_byte_for_prefix(encoded_signer_key)
        if prefix == _VersionByte.ED25519_PUBLIC_KEY:
            return cls.ed25519_public_key(encoded_signer_key)
        elif prefix == _VersionByte.PRE_AUTH_TX:
            return cls.pre_auth_tx(encoded_signer_key)
        elif prefix == _VersionByte.SHA256_HASH:
            return cls.sha256_hash(encoded_signer_key)
        elif prefix == _VersionByte.ED25519_SIGNED_PAYLOAD:
            return cls.ed25519_signed_payload(encoded_signer_key)
        else:
            raise ValueError(f"{prefix!r} is an unsupported version byte.")

    @classmethod
    def ed25519_public_key(cls, account_id: Union[str, bytes]) -> "SignerKey":
        """Create ED25519 PUBLIC KEY Signer from account id.

        :param account_id: account id
        :return: ED25519 PUBLIC KEY Signer
        :raises:
            :exc:`Ed25519PublicKeyInvalidError <stellar_sdk.exceptions.Ed25519PublicKeyInvalidError>`: if ``account_id``
            is not a valid ed25519 public key.
        """
        signer_key_type = SignerKeyType.SIGNER_KEY_TYPE_ED25519
        if isinstance(account_id, str):
            account_id = StrKey.decode_ed25519_public_key(account_id)
        return cls(signer_key=account_id, signer_key_type=signer_key_type)

    @classmethod
    def pre_auth_tx(cls, pre_auth_tx_hash: Union[str, bytes]) -> "SignerKey":
        """Create Pre AUTH TX Signer from the sha256 hash of a transaction,
        click `here <https://developers.stellar.org/docs/glossary/multisig/#pre-authorized-transaction>`__ for more information.

        :param pre_auth_tx_hash: The sha256 hash of a transaction.
        :return: Pre AUTH TX Signer
        """
        signer_key_type = SignerKeyType.SIGNER_KEY_TYPE_PRE_AUTH_TX
        if isinstance(pre_auth_tx_hash, str):
            pre_auth_tx_hash = StrKey.decode_pre_auth_tx(pre_auth_tx_hash)
        return cls(signer_key=pre_auth_tx_hash, signer_key_type=signer_key_type)

    @classmethod
    def sha256_hash(cls, sha256_hash: Union[str, bytes]) -> "SignerKey":
        """Create SHA256 HASH Signer from a sha256 hash of a preimage,
        click `here <https://developers.stellar.org/docs/glossary/multisig/#hashx>`__ for more information.

        :param sha256_hash: a sha256 hash of a preimage
        :return: SHA256 HASH Signer
        """
        signer_key_type = SignerKeyType.SIGNER_KEY_TYPE_HASH_X
        if isinstance(sha256_hash, str):
            sha256_hash = StrKey.decode_sha256_hash(sha256_hash)
        return cls(signer_key=sha256_hash, signer_key_type=signer_key_type)

    @classmethod
    def ed25519_signed_payload(
        cls, ed25519_signed_payload: Union[str, bytes, SignedPayloadSigner]
    ) -> "SignerKey":
        """Create ed25519 signed payload Signer from an ed25519 signed payload,
        click `here <https://github.com/stellar/stellar-protocol/blob/master/core/cap-0040.md>`__ for more information.

        :param ed25519_signed_payload: a sha256 hash of a preimage
        :return: ed25519 signed payload signer
        """
        signer_key_type = SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD
        if isinstance(ed25519_signed_payload, str):
            ed25519_signed_payload = StrKey.decode_ed25519_signed_payload(
                ed25519_signed_payload
            )
        if isinstance(ed25519_signed_payload, SignedPayloadSigner):
            decoded_account_id = StrKey.decode_ed25519_public_key(
                ed25519_signed_payload.account_id
            )
            payload_length = len(ed25519_signed_payload.payload)
            payload = ed25519_signed_payload.payload
            payload += b"\0" * ((4 - payload_length % 4) % 4)
            ed25519_signed_payload = (
                decoded_account_id
                + int.to_bytes(payload_length, length=4, byteorder="big")
                + payload
            )
        return cls(signer_key=ed25519_signed_payload, signer_key_type=signer_key_type)

    def to_signed_payload_signer(self) -> "SignedPayloadSigner":
        if self.signer_key_type != SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD:
            raise ValueError(
                f"{self.signer_key_type!r} is an unsupported signer key type, "
                f"it should be {SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD!r}."
            )

        payload_length = int.from_bytes(self.signer_key[32:36], byteorder="big")
        account_id = StrKey.encode_ed25519_public_key(self.signer_key[:32])
        payload = self.signer_key[36 : 36 + payload_length]
        return SignedPayloadSigner(account_id, payload)

    def to_xdr_object(self) -> stellar_xdr.SignerKey:
        """Returns the xdr object for this SignerKey object.

        :return: XDR Signer object
        """
        if self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_ED25519:
            return stellar_xdr.SignerKey(
                type=stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_ED25519,
                ed25519=stellar_xdr.Uint256(self.signer_key),
            )
        elif self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_PRE_AUTH_TX:
            return stellar_xdr.SignerKey(
                type=stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_PRE_AUTH_TX,
                pre_auth_tx=stellar_xdr.Uint256(self.signer_key),
            )
        elif (
            self.signer_key_type == SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD
        ):
            payload_length = int.from_bytes(self.signer_key[32:36], byteorder="big")
            ed25519_signed_payload = stellar_xdr.SignerKeyEd25519SignedPayload(
                ed25519=stellar_xdr.Uint256(self.signer_key[:32]),
                payload=self.signer_key[36 : 36 + payload_length],
            )
            return stellar_xdr.SignerKey(
                type=stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD,
                ed25519_signed_payload=ed25519_signed_payload,
            )
        else:
            return stellar_xdr.SignerKey(
                type=stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_HASH_X,
                hash_x=stellar_xdr.Uint256(self.signer_key),
            )

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.SignerKey) -> "SignerKey":
        """Create a :class:`SignerKey` from an XDR SignerKey object.

        :param xdr_object: The XDR SignerKey object.
        :return: A new :class:`SignerKey` object from the given XDR SignerKey object.
        """
        if xdr_object.type == stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_ED25519:
            assert xdr_object.ed25519 is not None
            return cls(
                xdr_object.ed25519.uint256, SignerKeyType.SIGNER_KEY_TYPE_ED25519
            )
        elif xdr_object.type == stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_PRE_AUTH_TX:
            assert xdr_object.pre_auth_tx is not None
            return cls(
                xdr_object.pre_auth_tx.uint256,
                SignerKeyType.SIGNER_KEY_TYPE_PRE_AUTH_TX,
            )
        elif xdr_object.type == stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_HASH_X:
            assert xdr_object.hash_x is not None
            return cls(xdr_object.hash_x.uint256, SignerKeyType.SIGNER_KEY_TYPE_HASH_X)
        elif (
            xdr_object.type
            == stellar_xdr.SignerKeyType.SIGNER_KEY_TYPE_ED25519_SIGNED_PAYLOAD
        ):
            assert xdr_object.ed25519_signed_payload is not None
            encoded_account_id = StrKey.encode_ed25519_public_key(
                xdr_object.ed25519_signed_payload.ed25519.uint256
            )
            payload = xdr_object.ed25519_signed_payload.payload
            ed25519_signed_payload = SignedPayloadSigner(encoded_account_id, payload)
            return cls.ed25519_signed_payload(
                ed25519_signed_payload=ed25519_signed_payload
            )
        else:
            raise ValueError(
                f"This is an unknown signer key type, please consider creating an issuer at {__issues__}."
            )

    def __hash__(self):
        return hash((self.signer_key, self.signer_key_type))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.signer_key == other.signer_key
            and self.signer_key_type == self.signer_key_type
        )

    def __str__(self):
        return f"<SignerKey [signer_key={self.signer_key}, signer_key_type={self.signer_key_type}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .manage_offer_success_result import ManageOfferSuccessResult
from .manage_sell_offer_result_code import ManageSellOfferResultCode

__all__ = ["ManageSellOfferResult"]


class ManageSellOfferResult:
    """
    XDR Source Code::

        union ManageSellOfferResult switch (ManageSellOfferResultCode code)
        {
        case MANAGE_SELL_OFFER_SUCCESS:
            ManageOfferSuccessResult success;
        case MANAGE_SELL_OFFER_MALFORMED:
        case MANAGE_SELL_OFFER_SELL_NO_TRUST:
        case MANAGE_SELL_OFFER_BUY_NO_TRUST:
        case MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED:
        case MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED:
        case MANAGE_SELL_OFFER_LINE_FULL:
        case MANAGE_SELL_OFFER_UNDERFUNDED:
        case MANAGE_SELL_OFFER_CROSS_SELF:
        case MANAGE_SELL_OFFER_SELL_NO_ISSUER:
        case MANAGE_SELL_OFFER_BUY_NO_ISSUER:
        case MANAGE_SELL_OFFER_NOT_FOUND:
        case MANAGE_SELL_OFFER_LOW_RESERVE:
            void;
        };
    """

    def __init__(
        self,
        code: ManageSellOfferResultCode,
        success: ManageOfferSuccessResult = None,
    ) -> None:
        self.code = code
        self.success = success

    def pack(self, packer: Packer) -> None:
        self.code.pack(packer)
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SUCCESS:
            if self.success is None:
                raise ValueError("success should not be None.")
            self.success.pack(packer)
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_MALFORMED:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SELL_NO_TRUST:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_BUY_NO_TRUST:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_LINE_FULL:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_UNDERFUNDED:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_CROSS_SELF:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SELL_NO_ISSUER:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_BUY_NO_ISSUER:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_NOT_FOUND:
            return
        if self.code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_LOW_RESERVE:
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ManageSellOfferResult:
        code = ManageSellOfferResultCode.unpack(unpacker)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SUCCESS:
            success = ManageOfferSuccessResult.unpack(unpacker)
            return cls(code=code, success=success)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_MALFORMED:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SELL_NO_TRUST:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_BUY_NO_TRUST:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SELL_NOT_AUTHORIZED:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_BUY_NOT_AUTHORIZED:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_LINE_FULL:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_UNDERFUNDED:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_CROSS_SELF:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_SELL_NO_ISSUER:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_BUY_NO_ISSUER:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_NOT_FOUND:
            return cls(code=code)
        if code == ManageSellOfferResultCode.MANAGE_SELL_OFFER_LOW_RESERVE:
            return cls(code=code)
        return cls(code=code)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ManageSellOfferResult:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ManageSellOfferResult:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.code,
                self.success,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.code == other.code and self.success == other.success

    def __str__(self):
        out = []
        out.append(f"code={self.code}")
        out.append(f"success={self.success}") if self.success is not None else None
        return f"<ManageSellOfferResult [{', '.join(out)}]>"



from decimal import Decimal
from enum import IntEnum
from typing import Optional, Sequence, Union

from .. import xdr as stellar_xdr
from ..asset import Asset
from ..keypair import Keypair
from ..muxed_account import MuxedAccount
from ..strkey import StrKey
from ..utils import raise_if_not_valid_amount, raise_if_not_valid_ed25519_public_key
from .operation import Operation

__all__ = ["ClaimPredicate", "Claimant", "CreateClaimableBalance"]


class ClaimPredicateType(IntEnum):
    """Currently supported claim predicate types."""

    CLAIM_PREDICATE_UNCONDITIONAL = 0
    CLAIM_PREDICATE_AND = 1
    CLAIM_PREDICATE_OR = 2
    CLAIM_PREDICATE_NOT = 3
    CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME = 4
    CLAIM_PREDICATE_BEFORE_RELATIVE_TIME = 5


class ClaimPredicateGroup:
    """Used to assemble the left and right values for and_predicates and or_predicates.

    :param left: The ClaimPredicate.
    :param right: The ClaimPredicate.
    """

    def __init__(self, left: "ClaimPredicate", right: "ClaimPredicate") -> None:
        self.left = left
        self.right = right

    def __hash__(self):
        return hash((self.left, self.right))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.left == other.left and self.right == other.right

    def __str__(self):
        return f"<ClaimPredicateGroup [left={self.left}, right={self.right}]>"


class ClaimPredicate:
    """The :class:`ClaimPredicate` object, which represents a ClaimPredicate on Stellar's network.

    **We do not recommend that you build it through the constructor, please use the helper function.**

    :param claim_predicate_type: Type of ClaimPredicate.
    :param and_predicates: The ClaimPredicates.
    :param or_predicates: The ClaimPredicates.
    :param not_predicate: The ClaimPredicate.
    :param abs_before: Unix epoch.
    :param rel_before: seconds since closeTime of the ledger in which the ClaimableBalanceEntry was created.
    """

    def __init__(
        self,
        claim_predicate_type: ClaimPredicateType,
        and_predicates: Optional[ClaimPredicateGroup],
        or_predicates: Optional[ClaimPredicateGroup],
        not_predicate: Optional["ClaimPredicate"],
        abs_before: Optional[int],
        rel_before: Optional[int],
    ) -> None:
        self.claim_predicate_type = claim_predicate_type
        self.and_predicates = and_predicates
        self.or_predicates = or_predicates
        self.not_predicate = not_predicate
        self.abs_before = abs_before
        self.rel_before = rel_before

    @classmethod
    def predicate_and(
        cls, left: "ClaimPredicate", right: "ClaimPredicate"
    ) -> "ClaimPredicate":
        """Returns an **and** claim predicate

        :param left: a ClaimPredicate.
        :param right: a ClaimPredicate.
        :return: an **and** claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_AND,
            and_predicates=ClaimPredicateGroup(left, right),
            or_predicates=None,
            not_predicate=None,
            abs_before=None,
            rel_before=None,
        )

    @classmethod
    def predicate_or(
        cls, left: "ClaimPredicate", right: "ClaimPredicate"
    ) -> "ClaimPredicate":
        """Returns an **or** claim predicate

        :param left: a ClaimPredicate.
        :param right: a ClaimPredicate.
        :return: an **or** claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_OR,
            and_predicates=None,
            or_predicates=ClaimPredicateGroup(left, right),
            not_predicate=None,
            abs_before=None,
            rel_before=None,
        )

    @classmethod
    def predicate_not(cls, predicate: "ClaimPredicate") -> "ClaimPredicate":
        """Returns a **not** claim predicate.

        :param predicate: a ClaimPredicate.
        :return: a **not** claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_NOT,
            and_predicates=None,
            or_predicates=None,
            not_predicate=predicate,
            abs_before=None,
            rel_before=None,
        )

    @classmethod
    def predicate_before_absolute_time(cls, abs_before: int) -> "ClaimPredicate":
        """Returns a **before_absolute_time** claim predicate.

        This predicate will be fulfilled if the closing time of the ledger that includes
        the :class:`CreateClaimableBalance` operation is less than this (absolute) Unix timestamp.

        :param abs_before: Unix epoch.
        :return: a **before_absolute_time** claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME,
            and_predicates=None,
            or_predicates=None,
            not_predicate=None,
            abs_before=abs_before,
            rel_before=None,
        )

    @classmethod
    def predicate_before_relative_time(cls, seconds: int) -> "ClaimPredicate":
        """Returns a **before_relative_time** claim predicate.

        This predicate will be fulfilled if the closing time of the ledger that
        includes the :class:`CreateClaimableBalance` operation plus this relative time delta (in seconds)
        is less than the current time.

        :param seconds: seconds since closeTime of the ledger in which the ClaimableBalanceEntry was created.
        :return: a **before_relative_time** claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME,
            and_predicates=None,
            or_predicates=None,
            not_predicate=None,
            abs_before=None,
            rel_before=seconds,
        )

    @classmethod
    def predicate_unconditional(cls) -> "ClaimPredicate":
        """Returns an **unconditional** claim predicate.

        :return: an **unconditional** claim predicate.
        """
        return cls(
            claim_predicate_type=ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL,
            and_predicates=None,
            or_predicates=None,
            not_predicate=None,
            abs_before=None,
            rel_before=None,
        )

    def to_xdr_object(self) -> stellar_xdr.ClaimPredicate:
        if (
            self.claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL
        ):
            return stellar_xdr.ClaimPredicate(
                stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL
            )
        elif (
            self.claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME
        ):
            assert self.abs_before is not None
            return stellar_xdr.ClaimPredicate(
                stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME,
                abs_before=stellar_xdr.Int64(self.abs_before),
            )
        elif (
            self.claim_predicate_type
            == ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME
        ):
            assert self.rel_before is not None
            return stellar_xdr.ClaimPredicate(
                stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME,
                rel_before=stellar_xdr.Int64(self.rel_before),
            )
        elif self.claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_NOT:
            assert self.not_predicate is not None
            return stellar_xdr.ClaimPredicate(
                stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_NOT,
                not_predicate=self.not_predicate.to_xdr_object(),
            )
        elif self.claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_AND:
            assert self.and_predicates is not None
            and_predicates = [
                self.and_predicates.left.to_xdr_object(),
                self.and_predicates.right.to_xdr_object(),
            ]
            return stellar_xdr.ClaimPredicate(
                stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_AND,
                and_predicates=and_predicates,
            )
        elif self.claim_predicate_type == ClaimPredicateType.CLAIM_PREDICATE_OR:
            assert self.or_predicates is not None
            or_predicates = [
                self.or_predicates.left.to_xdr_object(),
                self.or_predicates.right.to_xdr_object(),
            ]
            return stellar_xdr.ClaimPredicate(
                stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_OR,
                or_predicates=or_predicates,
            )
        else:
            raise ValueError(
                f"{self.claim_predicate_type} is not a valid ClaimPredicateType."
            )

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.ClaimPredicate
    ) -> "ClaimPredicate":
        claim_predicate_type = xdr_object.type
        if (
            claim_predicate_type
            == stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_UNCONDITIONAL
        ):
            return cls.predicate_unconditional()
        elif (
            claim_predicate_type
            == stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_BEFORE_ABSOLUTE_TIME
        ):
            assert xdr_object.abs_before is not None
            abs_before = xdr_object.abs_before.int64
            return cls.predicate_before_absolute_time(abs_before)
        elif (
            claim_predicate_type
            == stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_BEFORE_RELATIVE_TIME
        ):
            assert xdr_object.rel_before is not None
            rel_before = xdr_object.rel_before.int64
            return cls.predicate_before_relative_time(rel_before)
        elif claim_predicate_type == stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_NOT:
            not_predicate = xdr_object.not_predicate
            assert not_predicate is not None
            predicate = cls.from_xdr_object(not_predicate)
            return cls.predicate_not(predicate)
        elif claim_predicate_type == stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_AND:
            and_predicates = xdr_object.and_predicates
            assert and_predicates is not None
            left = cls.from_xdr_object(and_predicates[0])
            right = cls.from_xdr_object(and_predicates[1])
            return cls.predicate_and(left, right)
        elif claim_predicate_type == stellar_xdr.ClaimPredicateType.CLAIM_PREDICATE_OR:
            or_predicates = xdr_object.or_predicates
            assert or_predicates is not None
            left = cls.from_xdr_object(or_predicates[0])
            right = cls.from_xdr_object(or_predicates[1])
            return cls.predicate_or(left, right)
        else:
            raise ValueError(
                f"{claim_predicate_type} is an unsupported ClaimPredicateType."
            )

    def __hash__(self):
        return hash(
            (
                self.claim_predicate_type,
                self.and_predicates,
                self.or_predicates,
                self.not_predicate,
                self.abs_before,
                self.rel_before,
            )
        )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.claim_predicate_type == other.claim_predicate_type
            and self.and_predicates == other.and_predicates
            and self.or_predicates == other.or_predicates
            and self.not_predicate == other.not_predicate
            and self.abs_before == other.abs_before
            and self.rel_before == other.rel_before
        )

    def __str__(self):
        return (
            f"<ClaimPredicate [claim_predicate_type={self.claim_predicate_type}, "
            f"and_predicates={self.and_predicates}, "
            f"or_predicates={self.or_predicates}, "
            f"not_predicate={self.not_predicate}, "
            f"abs_before={self.abs_before}, "
            f"rel_before={self.rel_before}]>"
        )


class Claimant:
    """The :class:`Claimant` object represents a claimable balance claimant.

    :param destination: The destination account ID.
    :param predicate: The claim predicate. It is optional, it defaults to unconditional if none is specified.
    """

    def __init__(self, destination: str, predicate: ClaimPredicate = None) -> None:
        self.destination = destination
        if predicate is None:
            predicate = ClaimPredicate.predicate_unconditional()
        self.predicate: ClaimPredicate = predicate
        raise_if_not_valid_ed25519_public_key(self.destination, "destination")

    def to_xdr_object(self) -> stellar_xdr.Claimant:
        claimant_v0 = stellar_xdr.ClaimantV0(
            destination=Keypair.from_public_key(self.destination).xdr_account_id(),
            predicate=self.predicate.to_xdr_object(),
        )
        claimant = stellar_xdr.Claimant(
            stellar_xdr.ClaimantType.CLAIMANT_TYPE_V0, claimant_v0
        )
        return claimant

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Claimant) -> "Claimant":
        assert xdr_object.v0 is not None
        assert xdr_object.v0.destination.account_id.ed25519 is not None

        destination = StrKey.encode_ed25519_public_key(
            xdr_object.v0.destination.account_id.ed25519.uint256
        )
        predicate = ClaimPredicate.from_xdr_object(xdr_object.v0.predicate)
        return cls(destination=destination, predicate=predicate)

    def __hash__(self):
        return hash((self.destination, self.predicate))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.destination == other.destination and self.predicate == other.predicate
        )

    def __str__(self):
        return (
            f"<Claimant [destination={self.destination}, predicate={self.predicate}]>"
        )


class CreateClaimableBalance(Operation):
    """The :class:`CreateClaimableBalance` object, which represents a CreateClaimableBalance
    operation on Stellar's network.

    Creates a ClaimableBalanceEntry.
    See `Claimable Balance <https://developers.stellar.org/docs/glossary/claimable-balance/>`_
    for more information on parameters and usage.

    Threshold: Medium

    See `Create Claimable Balance <https://developers.stellar.org/docs/start/list-of-operations/#create-claimable-balance>`_ for more information.

    :param asset: The asset for the claimable balance.
    :param amount: the amount of the asset.
    :param claimants: A list of Claimants.
    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.CREATE_CLAIMABLE_BALANCE
    )

    def __init__(
        self,
        asset: Asset,
        amount: Union[str, Decimal],
        claimants: Sequence[Claimant],
        source: Optional[Union[MuxedAccount, str]] = None,
    ) -> None:
        super().__init__(source)
        self.asset: Asset = asset
        self.amount: str = str(amount)
        self.claimants: Sequence[Claimant] = claimants
        raise_if_not_valid_amount(self.amount, "amount")

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        asset = self.asset.to_xdr_object()
        amount = Operation.to_xdr_amount(self.amount)
        claimants = [claimant.to_xdr_object() for claimant in self.claimants]
        create_claimable_balance_op = stellar_xdr.CreateClaimableBalanceOp(
            asset=asset, amount=stellar_xdr.Int64(amount), claimants=claimants
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            create_claimable_balance_op=create_claimable_balance_op,
        )
        return body

    @classmethod
    def from_xdr_object(
        cls, xdr_object: stellar_xdr.Operation
    ) -> "CreateClaimableBalance":
        """Creates a :class:`CreateClaimableBalance` object from an XDR Operation
        object.
        """
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.create_claimable_balance_op is not None
        asset = Asset.from_xdr_object(xdr_object.body.create_claimable_balance_op.asset)
        amount = Operation.from_xdr_amount(
            xdr_object.body.create_claimable_balance_op.amount.int64
        )
        claimants = []
        for claimant_xdr_obj in xdr_object.body.create_claimable_balance_op.claimants:
            claimants.append(Claimant.from_xdr_object(claimant_xdr_obj))
        op = cls(asset=asset, amount=amount, claimants=claimants, source=source)
        return op

    def __str__(self):
        return (
            f"<CreateClaimableBalance [asset={self.asset}, amount={self.amount}, "
            f"claimants={self.claimants}, source={self.source}]>"
        )



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .int64 import Int64
from .ledger_header_ext import LedgerHeaderExt
from .stellar_value import StellarValue
from .uint32 import Uint32
from .uint64 import Uint64

__all__ = ["LedgerHeader"]


class LedgerHeader:
    """
    XDR Source Code::

        struct LedgerHeader
        {
            uint32 ledgerVersion;    // the protocol version of the ledger
            Hash previousLedgerHash; // hash of the previous ledger header
            StellarValue scpValue;   // what consensus agreed to
            Hash txSetResultHash;    // the TransactionResultSet that led to this ledger
            Hash bucketListHash;     // hash of the ledger state

            uint32 ledgerSeq; // sequence number of this ledger

            int64 totalCoins; // total number of stroops in existence.
                              // 10,000,000 stroops in 1 XLM

            int64 feePool;       // fees burned since last inflation run
            uint32 inflationSeq; // inflation sequence number

            uint64 idPool; // last used global ID, used for generating objects

            uint32 baseFee;     // base fee per operation in stroops
            uint32 baseReserve; // account base reserve in stroops

            uint32 maxTxSetSize; // maximum size a transaction set can be

            Hash skipList[4]; // hashes of ledgers in the past. allows you to jump back
                              // in time without walking the chain back ledger by ledger
                              // each slot contains the oldest ledger that is mod of
                              // either 50  5000  50000 or 500000 depending on index
                              // skipList[0] mod(50), skipList[1] mod(5000), etc

            // reserved for future use
            union switch (int v)
            {
            case 0:
                void;
            case 1:
                LedgerHeaderExtensionV1 v1;
            }
            ext;
        };
    """

    def __init__(
        self,
        ledger_version: Uint32,
        previous_ledger_hash: Hash,
        scp_value: StellarValue,
        tx_set_result_hash: Hash,
        bucket_list_hash: Hash,
        ledger_seq: Uint32,
        total_coins: Int64,
        fee_pool: Int64,
        inflation_seq: Uint32,
        id_pool: Uint64,
        base_fee: Uint32,
        base_reserve: Uint32,
        max_tx_set_size: Uint32,
        skip_list: List[Hash],
        ext: LedgerHeaderExt,
    ) -> None:
        _expect_length = 4
        if skip_list and len(skip_list) != _expect_length:
            raise ValueError(
                f"The length of `skip_list` should be {_expect_length}, but got {len(skip_list)}."
            )
        self.ledger_version = ledger_version
        self.previous_ledger_hash = previous_ledger_hash
        self.scp_value = scp_value
        self.tx_set_result_hash = tx_set_result_hash
        self.bucket_list_hash = bucket_list_hash
        self.ledger_seq = ledger_seq
        self.total_coins = total_coins
        self.fee_pool = fee_pool
        self.inflation_seq = inflation_seq
        self.id_pool = id_pool
        self.base_fee = base_fee
        self.base_reserve = base_reserve
        self.max_tx_set_size = max_tx_set_size
        self.skip_list = skip_list
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.ledger_version.pack(packer)
        self.previous_ledger_hash.pack(packer)
        self.scp_value.pack(packer)
        self.tx_set_result_hash.pack(packer)
        self.bucket_list_hash.pack(packer)
        self.ledger_seq.pack(packer)
        self.total_coins.pack(packer)
        self.fee_pool.pack(packer)
        self.inflation_seq.pack(packer)
        self.id_pool.pack(packer)
        self.base_fee.pack(packer)
        self.base_reserve.pack(packer)
        self.max_tx_set_size.pack(packer)
        for skip_list_item in self.skip_list:
            skip_list_item.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> LedgerHeader:
        ledger_version = Uint32.unpack(unpacker)
        previous_ledger_hash = Hash.unpack(unpacker)
        scp_value = StellarValue.unpack(unpacker)
        tx_set_result_hash = Hash.unpack(unpacker)
        bucket_list_hash = Hash.unpack(unpacker)
        ledger_seq = Uint32.unpack(unpacker)
        total_coins = Int64.unpack(unpacker)
        fee_pool = Int64.unpack(unpacker)
        inflation_seq = Uint32.unpack(unpacker)
        id_pool = Uint64.unpack(unpacker)
        base_fee = Uint32.unpack(unpacker)
        base_reserve = Uint32.unpack(unpacker)
        max_tx_set_size = Uint32.unpack(unpacker)
        length = 4
        skip_list = []
        for _ in range(length):
            skip_list.append(Hash.unpack(unpacker))
        ext = LedgerHeaderExt.unpack(unpacker)
        return cls(
            ledger_version=ledger_version,
            previous_ledger_hash=previous_ledger_hash,
            scp_value=scp_value,
            tx_set_result_hash=tx_set_result_hash,
            bucket_list_hash=bucket_list_hash,
            ledger_seq=ledger_seq,
            total_coins=total_coins,
            fee_pool=fee_pool,
            inflation_seq=inflation_seq,
            id_pool=id_pool,
            base_fee=base_fee,
            base_reserve=base_reserve,
            max_tx_set_size=max_tx_set_size,
            skip_list=skip_list,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> LedgerHeader:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> LedgerHeader:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.ledger_version,
                self.previous_ledger_hash,
                self.scp_value,
                self.tx_set_result_hash,
                self.bucket_list_hash,
                self.ledger_seq,
                self.total_coins,
                self.fee_pool,
                self.inflation_seq,
                self.id_pool,
                self.base_fee,
                self.base_reserve,
                self.max_tx_set_size,
                self.skip_list,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.ledger_version == other.ledger_version
            and self.previous_ledger_hash == other.previous_ledger_hash
            and self.scp_value == other.scp_value
            and self.tx_set_result_hash == other.tx_set_result_hash
            and self.bucket_list_hash == other.bucket_list_hash
            and self.ledger_seq == other.ledger_seq
            and self.total_coins == other.total_coins
            and self.fee_pool == other.fee_pool
            and self.inflation_seq == other.inflation_seq
            and self.id_pool == other.id_pool
            and self.base_fee == other.base_fee
            and self.base_reserve == other.base_reserve
            and self.max_tx_set_size == other.max_tx_set_size
            and self.skip_list == other.skip_list
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"ledger_version={self.ledger_version}",
            f"previous_ledger_hash={self.previous_ledger_hash}",
            f"scp_value={self.scp_value}",
            f"tx_set_result_hash={self.tx_set_result_hash}",
            f"bucket_list_hash={self.bucket_list_hash}",
            f"ledger_seq={self.ledger_seq}",
            f"total_coins={self.total_coins}",
            f"fee_pool={self.fee_pool}",
            f"inflation_seq={self.inflation_seq}",
            f"id_pool={self.id_pool}",
            f"base_fee={self.base_fee}",
            f"base_reserve={self.base_reserve}",
            f"max_tx_set_size={self.max_tx_set_size}",
            f"skip_list={self.skip_list}",
            f"ext={self.ext}",
        ]
        return f"<LedgerHeader [{', '.join(out)}]>"



from typing import List, Optional, Sequence, Union

from .. import xdr as stellar_xdr
from ..muxed_account import MuxedAccount
from .operation import Operation

__all__ = ["InvokeHostFunction"]


class InvokeHostFunction(Operation):
    """The :class:`InvokeHostFunction` object, which represents a InvokeHostFunction
    operation on Stellar's network.

    Threshold: Medium

    See `Interacting with Soroban via Stellar <https://soroban.stellar.org/docs/fundamentals-and-concepts/invoking-contracts-with-transactions>`_.

    :param host_function: The host function to invoke.
    :param auth: The authorizations required to execute the host function.
    :param source: The source account for the operation. Defaults to the transaction's source account.
    """

    _XDR_OPERATION_TYPE: stellar_xdr.OperationType = (
        stellar_xdr.OperationType.INVOKE_HOST_FUNCTION
    )

    def __init__(
        self,
        host_function: stellar_xdr.HostFunction,
        auth: Sequence[stellar_xdr.SorobanAuthorizationEntry] = None,
        source: Optional[Union[MuxedAccount, str]] = None,
    ):
        super().__init__(source)
        self.host_function = host_function
        self.auth: List[stellar_xdr.SorobanAuthorizationEntry] = (
            list(auth) if auth else []
        )

    def _to_operation_body(self) -> stellar_xdr.OperationBody:
        invoke_host_function_op = stellar_xdr.InvokeHostFunctionOp(
            host_function=self.host_function, auth=self.auth
        )
        body = stellar_xdr.OperationBody(
            type=self._XDR_OPERATION_TYPE,
            invoke_host_function_op=invoke_host_function_op,
        )
        return body

    @classmethod
    def from_xdr_object(cls, xdr_object: stellar_xdr.Operation) -> "InvokeHostFunction":
        """Creates a :class:`InvokeHostFunction` object from an XDR Operation object."""
        source = Operation.get_source_from_xdr_obj(xdr_object)
        assert xdr_object.body.invoke_host_function_op is not None
        host_function = xdr_object.body.invoke_host_function_op.host_function
        auth = xdr_object.body.invoke_host_function_op.auth
        return cls(
            host_function=host_function,
            auth=auth,
            source=source,
        )

    def __str__(self):
        return f"<InvokeHostFunction [host_function={self.host_function}, auth={self.auth}, source={self.source}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .sc_val import SCVal

__all__ = ["ContractEventV0"]


class ContractEventV0:
    """
    XDR Source Code::

        struct
                {
                    SCVal topics<>;
                    SCVal data;
                }
    """

    def __init__(
        self,
        topics: List[SCVal],
        data: SCVal,
    ) -> None:
        _expect_max_length = 4294967295
        if topics and len(topics) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `topics` should be {_expect_max_length}, but got {len(topics)}."
            )
        self.topics = topics
        self.data = data

    def pack(self, packer: Packer) -> None:
        packer.pack_uint(len(self.topics))
        for topics_item in self.topics:
            topics_item.pack(packer)
        self.data.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> ContractEventV0:
        length = unpacker.unpack_uint()
        topics = []
        for _ in range(length):
            topics.append(SCVal.unpack(unpacker))
        data = SCVal.unpack(unpacker)
        return cls(
            topics=topics,
            data=data,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> ContractEventV0:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> ContractEventV0:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.topics,
                self.data,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.topics == other.topics and self.data == other.data

    def __str__(self):
        out = [
            f"topics={self.topics}",
            f"data={self.data}",
        ]
        return f"<ContractEventV0 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .sc_meta_kind import SCMetaKind
from .sc_meta_v0 import SCMetaV0

__all__ = ["SCMetaEntry"]


class SCMetaEntry:
    """
    XDR Source Code::

        union SCMetaEntry switch (SCMetaKind kind)
        {
        case SC_META_V0:
            SCMetaV0 v0;
        };
    """

    def __init__(
        self,
        kind: SCMetaKind,
        v0: SCMetaV0 = None,
    ) -> None:
        self.kind = kind
        self.v0 = v0

    def pack(self, packer: Packer) -> None:
        self.kind.pack(packer)
        if self.kind == SCMetaKind.SC_META_V0:
            if self.v0 is None:
                raise ValueError("v0 should not be None.")
            self.v0.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCMetaEntry:
        kind = SCMetaKind.unpack(unpacker)
        if kind == SCMetaKind.SC_META_V0:
            v0 = SCMetaV0.unpack(unpacker)
            return cls(kind=kind, v0=v0)
        return cls(kind=kind)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCMetaEntry:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCMetaEntry:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.kind,
                self.v0,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.kind == other.kind and self.v0 == other.v0

    def __str__(self):
        out = []
        out.append(f"kind={self.kind}")
        out.append(f"v0={self.v0}") if self.v0 is not None else None
        return f"<SCMetaEntry [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .base import Integer
from .transaction_set_v1 import TransactionSetV1

__all__ = ["GeneralizedTransactionSet"]


class GeneralizedTransactionSet:
    """
    XDR Source Code::

        union GeneralizedTransactionSet switch (int v)
        {
        // We consider the legacy TransactionSet to be v0.
        case 1:
            TransactionSetV1 v1TxSet;
        };
    """

    def __init__(
        self,
        v: int,
        v1_tx_set: TransactionSetV1 = None,
    ) -> None:
        self.v = v
        self.v1_tx_set = v1_tx_set

    def pack(self, packer: Packer) -> None:
        Integer(self.v).pack(packer)
        if self.v == 1:
            if self.v1_tx_set is None:
                raise ValueError("v1_tx_set should not be None.")
            self.v1_tx_set.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> GeneralizedTransactionSet:
        v = Integer.unpack(unpacker)
        if v == 1:
            v1_tx_set = TransactionSetV1.unpack(unpacker)
            return cls(v=v, v1_tx_set=v1_tx_set)
        return cls(v=v)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> GeneralizedTransactionSet:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> GeneralizedTransactionSet:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.v,
                self.v1_tx_set,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return self.v == other.v and self.v1_tx_set == other.v1_tx_set

    def __str__(self):
        out = []
        out.append(f"v={self.v}")
        out.append(
            f"v1_tx_set={self.v1_tx_set}"
        ) if self.v1_tx_set is not None else None
        return f"<GeneralizedTransactionSet [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64

from xdrlib3 import Packer, Unpacker

from .envelope_type import EnvelopeType
from .fee_bump_transaction import FeeBumpTransaction
from .transaction import Transaction

__all__ = ["TransactionSignaturePayloadTaggedTransaction"]


class TransactionSignaturePayloadTaggedTransaction:
    """
    XDR Source Code::

        union switch (EnvelopeType type)
            {
            // Backwards Compatibility: Use ENVELOPE_TYPE_TX to sign ENVELOPE_TYPE_TX_V0
            case ENVELOPE_TYPE_TX:
                Transaction tx;
            case ENVELOPE_TYPE_TX_FEE_BUMP:
                FeeBumpTransaction feeBump;
            }
    """

    def __init__(
        self,
        type: EnvelopeType,
        tx: Transaction = None,
        fee_bump: FeeBumpTransaction = None,
    ) -> None:
        self.type = type
        self.tx = tx
        self.fee_bump = fee_bump

    def pack(self, packer: Packer) -> None:
        self.type.pack(packer)
        if self.type == EnvelopeType.ENVELOPE_TYPE_TX:
            if self.tx is None:
                raise ValueError("tx should not be None.")
            self.tx.pack(packer)
            return
        if self.type == EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP:
            if self.fee_bump is None:
                raise ValueError("fee_bump should not be None.")
            self.fee_bump.pack(packer)
            return

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> TransactionSignaturePayloadTaggedTransaction:
        type = EnvelopeType.unpack(unpacker)
        if type == EnvelopeType.ENVELOPE_TYPE_TX:
            tx = Transaction.unpack(unpacker)
            return cls(type=type, tx=tx)
        if type == EnvelopeType.ENVELOPE_TYPE_TX_FEE_BUMP:
            fee_bump = FeeBumpTransaction.unpack(unpacker)
            return cls(type=type, fee_bump=fee_bump)
        return cls(type=type)

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> TransactionSignaturePayloadTaggedTransaction:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> TransactionSignaturePayloadTaggedTransaction:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.type,
                self.tx,
                self.fee_bump,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.type == other.type
            and self.tx == other.tx
            and self.fee_bump == other.fee_bump
        )

    def __str__(self):
        out = []
        out.append(f"type={self.type}")
        out.append(f"tx={self.tx}") if self.tx is not None else None
        out.append(f"fee_bump={self.fee_bump}") if self.fee_bump is not None else None
        return f"<TransactionSignaturePayloadTaggedTransaction [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .account_entry_extension_v2_ext import AccountEntryExtensionV2Ext
from .constants import *
from .sponsorship_descriptor import SponsorshipDescriptor
from .uint32 import Uint32

__all__ = ["AccountEntryExtensionV2"]


class AccountEntryExtensionV2:
    """
    XDR Source Code::

        struct AccountEntryExtensionV2
        {
            uint32 numSponsored;
            uint32 numSponsoring;
            SponsorshipDescriptor signerSponsoringIDs<MAX_SIGNERS>;

            union switch (int v)
            {
            case 0:
                void;
            case 3:
                AccountEntryExtensionV3 v3;
            }
            ext;
        };
    """

    def __init__(
        self,
        num_sponsored: Uint32,
        num_sponsoring: Uint32,
        signer_sponsoring_i_ds: List[SponsorshipDescriptor],
        ext: AccountEntryExtensionV2Ext,
    ) -> None:
        _expect_max_length = MAX_SIGNERS
        if signer_sponsoring_i_ds and len(signer_sponsoring_i_ds) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `signer_sponsoring_i_ds` should be {_expect_max_length}, but got {len(signer_sponsoring_i_ds)}."
            )
        self.num_sponsored = num_sponsored
        self.num_sponsoring = num_sponsoring
        self.signer_sponsoring_i_ds = signer_sponsoring_i_ds
        self.ext = ext

    def pack(self, packer: Packer) -> None:
        self.num_sponsored.pack(packer)
        self.num_sponsoring.pack(packer)
        packer.pack_uint(len(self.signer_sponsoring_i_ds))
        for signer_sponsoring_i_ds_item in self.signer_sponsoring_i_ds:
            signer_sponsoring_i_ds_item.pack(packer)
        self.ext.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> AccountEntryExtensionV2:
        num_sponsored = Uint32.unpack(unpacker)
        num_sponsoring = Uint32.unpack(unpacker)
        length = unpacker.unpack_uint()
        signer_sponsoring_i_ds = []
        for _ in range(length):
            signer_sponsoring_i_ds.append(SponsorshipDescriptor.unpack(unpacker))
        ext = AccountEntryExtensionV2Ext.unpack(unpacker)
        return cls(
            num_sponsored=num_sponsored,
            num_sponsoring=num_sponsoring,
            signer_sponsoring_i_ds=signer_sponsoring_i_ds,
            ext=ext,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> AccountEntryExtensionV2:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> AccountEntryExtensionV2:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.num_sponsored,
                self.num_sponsoring,
                self.signer_sponsoring_i_ds,
                self.ext,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.num_sponsored == other.num_sponsored
            and self.num_sponsoring == other.num_sponsoring
            and self.signer_sponsoring_i_ds == other.signer_sponsoring_i_ds
            and self.ext == other.ext
        )

    def __str__(self):
        out = [
            f"num_sponsored={self.num_sponsored}",
            f"num_sponsoring={self.num_sponsoring}",
            f"signer_sponsoring_i_ds={self.signer_sponsoring_i_ds}",
            f"ext={self.ext}",
        ]
        return f"<AccountEntryExtensionV2 [{', '.join(out)}]>"



# This is an automatically generated file.
# DO NOT EDIT or your changes may be overwritten
from __future__ import annotations

import base64
from typing import List

from xdrlib3 import Packer, Unpacker

from .hash import Hash
from .value import Value

__all__ = ["SCPNomination"]


class SCPNomination:
    """
    XDR Source Code::

        struct SCPNomination
        {
            Hash quorumSetHash; // D
            Value votes<>;      // X
            Value accepted<>;   // Y
        };
    """

    def __init__(
        self,
        quorum_set_hash: Hash,
        votes: List[Value],
        accepted: List[Value],
    ) -> None:
        _expect_max_length = 4294967295
        if votes and len(votes) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `votes` should be {_expect_max_length}, but got {len(votes)}."
            )
        _expect_max_length = 4294967295
        if accepted and len(accepted) > _expect_max_length:
            raise ValueError(
                f"The maximum length of `accepted` should be {_expect_max_length}, but got {len(accepted)}."
            )
        self.quorum_set_hash = quorum_set_hash
        self.votes = votes
        self.accepted = accepted

    def pack(self, packer: Packer) -> None:
        self.quorum_set_hash.pack(packer)
        packer.pack_uint(len(self.votes))
        for votes_item in self.votes:
            votes_item.pack(packer)
        packer.pack_uint(len(self.accepted))
        for accepted_item in self.accepted:
            accepted_item.pack(packer)

    @classmethod
    def unpack(cls, unpacker: Unpacker) -> SCPNomination:
        quorum_set_hash = Hash.unpack(unpacker)
        length = unpacker.unpack_uint()
        votes = []
        for _ in range(length):
            votes.append(Value.unpack(unpacker))
        length = unpacker.unpack_uint()
        accepted = []
        for _ in range(length):
            accepted.append(Value.unpack(unpacker))
        return cls(
            quorum_set_hash=quorum_set_hash,
            votes=votes,
            accepted=accepted,
        )

    def to_xdr_bytes(self) -> bytes:
        packer = Packer()
        self.pack(packer)
        return packer.get_buffer()

    @classmethod
    def from_xdr_bytes(cls, xdr: bytes) -> SCPNomination:
        unpacker = Unpacker(xdr)
        return cls.unpack(unpacker)

    def to_xdr(self) -> str:
        xdr_bytes = self.to_xdr_bytes()
        return base64.b64encode(xdr_bytes).decode()

    @classmethod
    def from_xdr(cls, xdr: str) -> SCPNomination:
        xdr_bytes = base64.b64decode(xdr.encode())
        return cls.from_xdr_bytes(xdr_bytes)

    def __hash__(self):
        return hash(
            (
                self.quorum_set_hash,
                self.votes,
                self.accepted,
            )
        )

    def __eq__(self, other: object):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return (
            self.quorum_set_hash == other.quorum_set_hash
            and self.votes == other.votes
            and self.accepted == other.accepted
        )

    def __str__(self):
        out = [
            f"quorum_set_hash={self.quorum_set_hash}",
            f"votes={self.votes}",
            f"accepted={self.accepted}",
        ]
        return f"<SCPNomination [{', '.join(out)}]>"
